# MIT License
#
# Copyright (c) 2020 Jonathan Zernik
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
import queue
import socket
import threading

import socks

from squeaknode.core.peer_address import Network
from squeaknode.core.peer_address import PeerAddress


SOCKET_CONNECT_TIMEOUT = 30


logger = logging.getLogger(__name__)


class PeerClient(object):
    """Creates outgoing connections to other peers in the network.
    """

    def __init__(self, peer_handler, tor_proxy_ip, tor_proxy_port):
        self.peer_handler = peer_handler
        self.tor_proxy_ip = tor_proxy_ip
        self.tor_proxy_port = tor_proxy_port

    def connect_address(self, address: PeerAddress):
        logger.info('Making connection to {}'.format(address))
        result_queue: queue.Queue = queue.Queue()
        threading.Thread(
            target=self.make_connection,
            args=(address, result_queue,),
        ).start()

        # Wait for connect result from the queue.
        connect_result = result_queue.get()
        logger.debug("connect_result: {}".format(connect_result))
        if connect_result.failure is not None:
            raise connect_result.failure

    def connect_address_async(self, address: PeerAddress):
        logger.info('Making connection async to {}'.format(address))
        result_queue: queue.Queue = queue.Queue()
        threading.Thread(
            target=self.make_connection,
            args=(address, result_queue,),
        ).start()

    def make_connection(self, address: PeerAddress, result_queue: queue.Queue):
        logger.info('Conecting to address: {}'.format(address))
        try:
            peer_socket = self.get_socket(address)
            peer_socket.settimeout(SOCKET_CONNECT_TIMEOUT)
            connect_address = (address.host, address.port)
            peer_socket.connect(connect_address)
            peer_socket.setblocking(True)
            self.handle_connection(
                peer_socket,
                address,
                result_queue,
            )
        except Exception as e:
            logger.exception('Failed to connect to {}'.format(address))
            result_queue.put(ConnectPeerResult.from_failure(e))

    def handle_connection(
            self,
            peer_socket: socket.socket,
            peer_address: PeerAddress,
            result_queue: queue.Queue,
    ):
        """Handle a newly connected peer socket."""
        self.peer_handler.handle_connection(
            peer_socket,
            peer_address,
            outgoing=True,
            result_queue=result_queue,
        )

    def get_socket(self, address: PeerAddress):
        if address.network not in [
                Network.IPV4,
                Network.IPV6,
                Network.TORV3,
        ]:
            raise Exception("Unsupported network: {}".format(address.network))
        if address.network == Network.TORV3 and self.tor_proxy_ip is None:
            raise Exception(
                "Unable to connect to tor address without tor proxy ip configured.")
        if address.network == Network.TORV3 and self.tor_proxy_port is None:
            raise Exception(
                "Unable to connect to tor address without tor proxy port configured.")
        if address.network == Network.TORV3:
            s = socks.socksocket()  # Same API as socket.socket in the standard lib
            s.set_proxy(socks.SOCKS5, self.tor_proxy_ip, self.tor_proxy_port)
            return s
        return socket.socket()


class ConnectPeerResult(object):
    """Result of a connect peer attempt.
    """

    def __init__(
            self,
            success: PeerAddress = None,
            failure: Exception = None,
    ):
        self.success = success
        self.failure = failure

    @classmethod
    def from_success(cls, success):
        return cls(success=success)

    @classmethod
    def from_failure(cls, failure):
        return cls(failure=failure)

    def __repr__(self):
        return "ConnectPeerResult: \
        success: {} \
        failure: {}".format(
            self.success,
            self.failure,
        )



# MIT License
#
# Copyright (c) 2020 Jonathan Zernik
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
import threading
from typing import List
from typing import Optional

from squeaknode.core.twitter_account_entry import TwitterAccountEntry
from squeaknode.node.squeak_controller import SqueakController
from squeaknode.twitter.twitter_stream import TwitterStream


logger = logging.getLogger(__name__)


class TwitterForwarder:

    def __init__(
            self,
            retry_s: int,
    ):
        self.retry_s = retry_s
        self.lock = threading.Lock()
        self.current_task: Optional[TwitterForwarderTask] = None

    def start_processing(self, squeak_controller: SqueakController):
        with self.lock:
            if self.current_task is not None:
                self.current_task.stop_processing()
            self.current_task = TwitterForwarderTask(
                squeak_controller,
                self.retry_s,
            )
            self.current_task.start_processing()

    def stop_processing(self):
        with self.lock:
            if self.current_task is not None:
                self.current_task.stop_processing()


class TwitterForwarderTask:

    def __init__(
        self,
        squeak_controller: SqueakController,
        retry_s: int,
    ):
        self.squeak_controller = squeak_controller
        self.retry_s = retry_s
        self.stopped = threading.Event()
        self.tweet_stream = None

    def start_processing(self):
        logger.info("Starting twitter forwarder task.")
        threading.Thread(
            target=self.process_forward_tweets,
            daemon=True,
        ).start()

    def stop_processing(self):
        logger.info("Stopping twitter forwarder task.")
        self.stopped.set()
        if self.tweet_stream is not None:
            self.tweet_stream.cancel_fn()

    def process_forward_tweets(self):
        while not self.stopped.is_set():
            try:
                bearer_token = self.get_bearer_token()
                handles = self.get_twitter_handles()
                if not bearer_token:
                    return
                if not handles:
                    return
                logger.info("Starting forward tweets with bearer token: {} and twitter handles: {}".format(
                    bearer_token,
                    handles,
                ))
                twitter_stream = TwitterStream(bearer_token, handles)
                self.tweet_stream = twitter_stream.get_tweets()
                if self.stopped.is_set():
                    return
                for tweet in self.tweet_stream.result_stream:
                    self.handle_tweet(tweet)
            # TODO: use more specific error.
            except Exception:
                logger.exception(
                    "Unable to subscribe tweet stream. Retrying in {} seconds...".format(
                        self.retry_s,
                    ),
                )
                self.stopped.wait(self.retry_s)

    def get_bearer_token(self) -> str:
        return self.squeak_controller.get_twitter_bearer_token() or ''

    def get_twitter_handles(self) -> List[str]:
        twitter_accounts = self.squeak_controller.get_twitter_accounts()
        handles = [account.handle for account in twitter_accounts]
        return handles

    def is_tweet_a_match(self, tweet: dict, account: TwitterAccountEntry) -> bool:
        for rule in tweet['matching_rules']:
            if rule['tag'] == account.handle:
                return True
        return False

    def forward_tweet(self, tweet: dict, account: TwitterAccountEntry) -> None:
        self.squeak_controller.make_squeak(
            profile_id=account.profile_id,
            content_str=tweet['data']['text'],
            replyto_hash=None,
        )

    def handle_tweet(self, tweet: dict):
        logger.info(
            "Got tweet: {}".format(tweet))
        twitter_accounts = self.squeak_controller.get_twitter_accounts()
        for account in twitter_accounts:
            if self.is_tweet_a_match(tweet, account):
                self.forward_tweet(tweet, account)



# MIT License
#
# Copyright (c) 2020 Jonathan Zernik
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
from typing import Optional

from squeak.core import CSqueak

from squeaknode.core.interests import squeak_matches_interest
from squeaknode.core.offer import Offer
from squeaknode.core.peer_address import PeerAddress
from squeaknode.node.squeak_controller import SqueakController


logger = logging.getLogger(__name__)


class DownloadHandler:
    """Handles received squeaks.
    """

    def __init__(self, squeak_controller: SqueakController):
        self.squeak_controller = squeak_controller

    def handle_squeak(self, squeak: CSqueak):
        """Handle a new received squeak.

        """
        # Try saving squeak as active download
        saved_squeak_hash = self.save_active_download_squeak(squeak)
        if saved_squeak_hash is None:
            saved_squeak_hash = self.save_followed_squeak(squeak)
        if saved_squeak_hash is not None:
            self.squeak_controller.request_offers(saved_squeak_hash)

    def handle_offer(self, offer: Offer, peer_address: PeerAddress):
        """Handle a new received offer.

        """
        received_offer_id = self.squeak_controller.save_received_offer(
            offer, peer_address)
        if received_offer_id is None:
            return
        counter = self.squeak_controller.get_download_offer_counter(offer)
        if counter is not None:
            counter.increment()

    def save_active_download_squeak(self, squeak: CSqueak) -> Optional[bytes]:
        """Save the given squeak as a result of an active download.

        Returns:
          bytes: the hash of the saved squeak.
        """
        counter = self.squeak_controller.get_download_squeak_counter(squeak)
        if counter is None:
            return None
        saved_squeak_hash = self.squeak_controller.save_squeak(squeak)
        if saved_squeak_hash is None:
            return None
        counter.increment()
        return saved_squeak_hash

    def save_followed_squeak(self, squeak: CSqueak) -> Optional[bytes]:
        """Save the given squeak because it matches the followed
        interest criteria.

        Returns:
          bytes: the hash of the saved squeak.
        """
        if not self.squeak_matches_interest(squeak):
            return None
        return self.squeak_controller.save_squeak(squeak)

    def squeak_matches_interest(self, squeak: CSqueak) -> bool:
        locator = self.squeak_controller.get_interested_locator()
        for interest in locator.vInterested:
            if squeak_matches_interest(squeak, interest) \
               and self.squeak_controller.squeak_in_limit_of_interest(squeak, interest):
                return True
        return False



# MIT License
#
# Copyright (c) 2020 Jonathan Zernik
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
import threading

from squeak.messages import msg_addr
from squeak.messages import msg_getaddr
from squeak.messages import msg_getdata
from squeak.messages import msg_inv
from squeak.messages import msg_notfound
from squeak.messages import msg_ping
from squeak.messages import msg_pong
from squeak.messages import MSG_SECRET_KEY
from squeak.messages import MSG_SQUEAK
from squeak.messages import msg_squeak
from squeak.net import CInterested
from squeak.net import CInv

from squeaknode.core.crypto import generate_ping_nonce
from squeaknode.core.offer import Offer
from squeaknode.network.download_handler import DownloadHandler
from squeaknode.network.peer import Peer
from squeaknode.node.squeak_controller import SqueakController


logger = logging.getLogger(__name__)


EMPTY_HASH = b'\x00' * 32
HANDSHAKE_TIMEOUT = 30
PING_TIMEOUT = 60
PONG_TIMEOUT = 30


class Connection(object):
    """Handles lifecycle of a connection to a peer.
    """

    def __init__(self, peer: Peer, squeak_controller: SqueakController):
        self.peer = peer
        self.squeak_controller = squeak_controller
        self.handshake_timer = HandshakeTimer(self)
        self.ping_timer = PingTimer(self)
        self.pong_timer = PongTimer(self)
        self.download_handler = DownloadHandler(self.squeak_controller)

    def handshake(self):
        """Do a handshake with a peer.
        """
        self.handshake_timer.start_timer()

        if self.peer.outgoing:
            self.peer.send_version()
        self.peer.receive_version()
        if not self.peer.outgoing:
            self.peer.send_version()

        self.peer.set_connected()
        self.handshake_timer.cancel()
        logger.debug("HANDSHAKE COMPLETE-----------")

    def shutdown(self):
        logger.info("Peer shutting down...")
        self.peer.stop()
        self.handshake_timer.cancel()
        self.ping_timer.cancel()
        self.pong_timer.cancel()

    def handle_connection(self):
        try:
            self.initial_sync()
            self.handle_msgs()
        except Exception:
            logger.exception("Error in handle_connection")
        finally:
            self.shutdown()
            # self._stopped.set()

    def initial_sync(self):
        self.send_ping()
        self.update_addrs()
        self.update_subscription()

    def send_ping(self):
        ping_msg = msg_ping()
        ping_msg.nonce = generate_ping_nonce()
        self.pong_timer.start_timer(ping_msg.nonce)
        self.peer.send_msg(ping_msg)

    def start_ping_timer(self):
        self.ping_timer.start_timer()

    def update_subscription(self):
        locator = self.squeak_controller.get_interested_locator()
        self.peer.update_local_subscription(locator)

    def update_addrs(self):
        getaddr_msg = msg_getaddr()
        self.peer.send_msg(getaddr_msg)

    def handle_msgs(self):
        """Handles messages from the peer if there are any available.

        This method blocks when the peer has not sent any messages.
        """
        msg = self.peer.recv_msg()
        while msg is not None:
            self.handle_peer_message(msg)
            msg = self.peer.recv_msg()
        logger.info("Finished handle_msgs")

    def handle_peer_message(self, msg):
        """Handle messages from a peer with completed handshake."""
        if msg.command == b'ping':
            self.handle_ping(msg)
        elif msg.command == b'pong':
            self.handle_pong(msg)
        elif msg.command == b'addr':
            self.handle_addr(msg)
        elif msg.command == b'getaddr':
            self.handle_getaddr(msg)
        elif msg.command == b'inv':
            self.handle_inv(msg)
        elif msg.command == b'getsqueaks':
            self.handle_getsqueaks(msg)
        elif msg.command == b'squeak':
            self.handle_squeak(msg)
        elif msg.command == b'getdata':
            self.handle_getdata(msg)
        elif msg.command == b'notfound':
            self.handle_notfound(msg)
        elif msg.command == b'offer':
            self.handle_offer(msg)
        elif msg.command == b'secretkey':
            self.handle_secret_key(msg)
        elif msg.command == b'subscribe':
            self.handle_subscribe(msg)
        else:
            raise Exception("Unrecognized message: {}".format(
                msg.command
            ))

    def handle_ping(self, msg):
        nonce = msg.nonce
        pong = msg_pong()
        pong.nonce = nonce
        self.peer.set_last_recv_ping()
        self.peer.send_msg(pong)

    def handle_pong(self, msg):
        nonce = msg.nonce
        self.pong_timer.stop_timer(nonce)

    def handle_addr(self, msg):
        # TODO: Save new address in table.
        for addr in msg.addrs:
            pass

    def handle_getaddr(self, msg):
        # TODO: Get known peers from table in database.
        addr_msg = msg_addr(addrs=[])
        self.peer.send_msg(addr_msg)

    def handle_inv(self, msg):
        invs = msg.inv
        unknown_squeak_invs = [
            inv for inv in invs
            if inv.type == MSG_SQUEAK
            and self.squeak_controller.get_squeak(inv.hash) is None
        ]
        unknown_secret_key_invs = [
            inv for inv in invs
            if inv.type == MSG_SECRET_KEY
            and self.squeak_controller.get_squeak(inv.hash) is not None
            and self.squeak_controller.get_squeak_secret_key(inv.hash) is None
        ]
        unknown_invs = unknown_squeak_invs + unknown_secret_key_invs
        if unknown_invs:
            getdata_msg = msg_getdata(inv=unknown_invs)
            self.peer.send_msg(getdata_msg)

    def handle_getdata(self, msg):
        invs = msg.inv
        not_found = []
        for inv in invs:
            reply_msg = self._get_inv_reply(inv)
            if reply_msg is not None:
                self.peer.send_msg(reply_msg)
            else:
                not_found.append(inv)
        if not_found:
            notfound_msg = msg_notfound(inv=not_found)
            self.peer.send_msg(notfound_msg)

    def handle_notfound(self, msg):
        pass

    def handle_getsqueaks(self, msg):
        self._send_reply_invs(msg.locator)

    def handle_squeak(self, msg):
        squeak = msg.squeak
        self.download_handler.handle_squeak(squeak)

    def handle_offer(self, msg):
        offer = Offer(
            squeak_hash=msg.hashSqk,
            nonce=msg.nonce,
            payment_request=msg.strPaymentInfo.decode('utf-8'),
            host=msg.host.decode('utf-8'),
            port=msg.port,
        )
        self.download_handler.handle_offer(
            offer,
            self.peer.remote_address,
        )

    def handle_secret_key(self, msg):
        self.squeak_controller.unlock_squeak(
            msg.hashSqk,
            msg.secretKey,
        )

    def handle_subscribe(self, msg):
        if msg.protover < 60003:
            self._send_reply_invs(msg.locator)
        self.peer.set_remote_subscription(msg.locator)

    def _send_reply_invs(self, locator):
        for interest in locator.vInterested:
            squeak_hashes = self._get_local_squeaks(interest)
            secret_key_hashes = self._get_local_secret_keys(interest)
            squeak_invs = [
                CInv(type=MSG_SQUEAK, hash=squeak_hash)
                for squeak_hash in squeak_hashes]
            secret_key_invs = [
                CInv(type=MSG_SECRET_KEY, hash=squeak_hash)
                for squeak_hash in secret_key_hashes]
            invs = squeak_invs + secret_key_invs
            if invs:
                inv_msg = msg_inv(inv=invs)
                self.peer.send_msg(inv_msg)

    def _get_local_squeaks(self, interest: CInterested):
        min_block = interest.nMinBlockHeight if interest.nMinBlockHeight != -1 else None
        max_block = interest.nMaxBlockHeight if interest.nMaxBlockHeight != -1 else None
        reply_to_hash = interest.hashReplySqk if interest.hashReplySqk != EMPTY_HASH else None
        return self.squeak_controller.lookup_squeaks(
            addresses=[str(address) for address in interest.addresses],
            min_block=min_block,
            max_block=max_block,
            reply_to_hash=reply_to_hash,
        )

    def _get_local_secret_keys(self, interest: CInterested):
        min_block = interest.nMinBlockHeight if interest.nMinBlockHeight != -1 else None
        max_block = interest.nMaxBlockHeight if interest.nMaxBlockHeight != -1 else None
        reply_to_hash = interest.hashReplySqk if interest.hashReplySqk != EMPTY_HASH else None
        return self.squeak_controller.lookup_secret_keys(
            addresses=[str(address) for address in interest.addresses],
            min_block=min_block,
            max_block=max_block,
            reply_to_hash=reply_to_hash,
        )

    def _get_inv_reply(self, inv):
        if inv.type == MSG_SQUEAK:
            return self._get_inv_reply_for_squeak(inv)
        if inv.type == MSG_SECRET_KEY:
            return self._get_inv_reply_for_secret_key(inv)

    def _get_inv_reply_for_squeak(self, inv):
        squeak = self.squeak_controller.get_squeak(inv.hash)
        if squeak is not None:
            return msg_squeak(squeak=squeak)

    def _get_inv_reply_for_secret_key(self, inv):
        resp = self.squeak_controller.get_secret_key_reply(
            inv.hash,
            self.peer.remote_address,
        )
        if resp is None:
            return None
        return resp.get_msg()


class HandshakeTimer:
    """Stop the peer if handshake is not complete before timeout.
    """

    def __init__(self, connection: Connection):
        self.connection = connection
        self.timer = None
        self._lock = threading.Lock()

    def start_timer(self):
        with self._lock:
            self.timer = threading.Timer(
                HANDSHAKE_TIMEOUT,
                self.shutdown,
            )
            self.timer.name = "handshake_timer_thread_{}".format(
                self.connection.peer)
            self.timer.start()

    def cancel(self):
        logger.debug("Cancelling handshake timer.")
        with self._lock:
            if self.timer:
                self.timer.cancel()

    def shutdown(self):
        logger.info("Shutdown connection triggered by handshake timer.")
        self.connection.shutdown()


class PingTimer:
    """Send a ping message when the timer expires.
    """

    def __init__(self, connection: Connection):
        self.connection = connection
        self.timer = None
        self._lock = threading.Lock()

    def start_timer(self):
        logger.debug("Starting ping timer.")
        with self._lock:
            # Cancel the existing timer.
            if self.timer:
                self.timer.cancel()

            # Start a new timer.
            self.timer = threading.Timer(
                PING_TIMEOUT,
                self.send_ping,
            )
            self.timer.name = "ping_timer_thread_{}".format(
                self.connection.peer)
            self.timer.start()

    def cancel(self):
        logger.debug("Cancelling ping timer.")
        with self._lock:
            if self.timer:
                self.timer.cancel()

    def send_ping(self):
        logger.debug("Sending ping triggered by timer.")
        self.connection.send_ping()


class PongTimer:
    """Shut down the connection when the timer expires.
    """

    def __init__(self, connection: Connection):
        self.connection = connection
        self.timer = None
        self.expected_nonce = None
        self._lock = threading.Lock()

    def start_timer(self, nonce):
        logger.debug("Starting pong timer.")
        with self._lock:
            # Cancel the existing timer.
            if self.timer:
                return

            # Start a new timer.
            self.expected_nonce = nonce
            self.timer = threading.Timer(
                PONG_TIMEOUT,
                self.shutdown,
            )
            self.timer.name = "pong_timer_thread_{}".format(
                self.connection.peer)
            self.timer.start()

    def stop_timer(self, nonce):
        logger.debug("Stopping pong timer.")
        with self._lock:
            if nonce != self.expected_nonce:
                self.shutdown()

            # Cancel the existing timer.
            if self.timer:
                self.timer.cancel()
                self.timer = None
                self.expected_nonce = None

            # Start a new ping timer.
            self.start_ping_timer()

    def cancel(self):
        logger.debug("Cancelling pong timer.")
        with self._lock:
            if self.timer:
                self.timer.cancel()

    def shutdown(self):
        logger.info("Shutdown connection triggered by pong timer.")
        self.connection.shutdown()

    def start_ping_timer(self):
        logger.debug("Starting ping timer triggered by pong response.")
        self.connection.start_ping_timer()



# MIT License
#
# Copyright (c) 2020 Jonathan Zernik
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
from abc import ABC
from abc import abstractmethod

from squeaknode.lightning.info import Info
from squeaknode.lightning.invoice import Invoice
from squeaknode.lightning.invoice_stream import InvoiceStream
from squeaknode.lightning.pay_req import PayReq
from squeaknode.lightning.payment import Payment

logger = logging.getLogger(__name__)


class LightningClient(ABC):

    @abstractmethod
    def get_info(self) -> Info:
        """Get info about the lightning node.

        Returns:
            Info: an object containing information about the node.

        Raises:
            LightningRequestError: If the request fails.
        """

    @abstractmethod
    def create_invoice(self, preimage: bytes, amount_msat: int) -> Invoice:
        """Add a new invoice.

        Args:
            preimage: The lightning invoice preimage.
            amount_msat: The amount of the invoice in msats.

        Returns:
            Invoice: an object representing a lightning invoice.

        Raises:
            LightningRequestError: If the request fails.
        """

    @abstractmethod
    def decode_pay_req(self, payment_request: str) -> PayReq:
        """Get the decoded payment request.

        Args:
            payment_request: The payment request as a string.

        Returns:
            PayReq: an object representing the payment request.

        Raises:
            LightningRequestError: If the request fails.
        """

    @abstractmethod
    def pay_invoice(self, payment_request: str) -> Payment:
        """Pay an invoice with a given payment_request.

        Args:
            payment_request: The payment request as a string.

        Returns:
            Payment: The payment result.

        args:
        payment_request -- the payment_request as a string
        """

    @abstractmethod
    def subscribe_invoices(self, settle_index: int) -> InvoiceStream:
        """Get a stream of settled invoices for received payments.
        # TODO: use map function to convert type of items in stream.

        Args:
            settle_index: The settle index from which to start streaming.

        Returns:
            InvoiceStream: an object containing the stream of invoices.

        Raises:
            LightningRequestError: If the request fails.
        """



# MIT License
#
# Copyright (c) 2020 Jonathan Zernik
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
from typing import Callable
from typing import Optional
from typing import Tuple

import grpc
from bitcoin.core import CBlockHeader
from squeak.core import CSqueak
from squeak.core.signing import CSigningKey

from squeaknode.bitcoin.bitcoin_client import BitcoinClient
from squeaknode.core.exception import InvoiceSubscriptionError
from squeaknode.core.lightning_address import LightningAddressHostPort
from squeaknode.core.offer import Offer
from squeaknode.core.peer_address import PeerAddress
from squeaknode.core.received_offer import ReceivedOffer
from squeaknode.core.received_payment import ReceivedPayment
from squeaknode.core.received_payment_stream import ReceivedPaymentsStream
from squeaknode.core.secret_keys import add_tweak
from squeaknode.core.secret_keys import generate_tweak
from squeaknode.core.secret_keys import subtract_tweak
from squeaknode.core.sent_offer import SentOffer
from squeaknode.core.sent_payment import SentPayment
from squeaknode.core.squeak_profile import SqueakProfile
from squeaknode.core.squeaks import check_squeak
from squeaknode.core.squeaks import get_decrypted_content
from squeaknode.core.squeaks import get_hash
from squeaknode.core.squeaks import get_payment_point_of_secret_key
from squeaknode.core.squeaks import make_squeak_with_block
from squeaknode.lightning.lightning_client import LightningClient


logger = logging.getLogger(__name__)


class SqueakCore:
    def __init__(
        self,
        bitcoin_client: BitcoinClient,
        lightning_client: LightningClient,
    ):
        self.bitcoin_client = bitcoin_client
        self.lightning_client = lightning_client

    def make_squeak(self, signing_profile: SqueakProfile, content_str: str, replyto_hash: Optional[bytes] = None) -> Tuple[CSqueak, bytes]:
        """Create a new squeak.

        Args:
            signing_profile: The profile of the author of the squeak.
            content_str: The content of the squeak as a string.
            replyto_hash: The hash of the squeak to which this one is replying.

        Returns:
            Tuple[CSqueak, bytes]: the squeak that was created together
        with its decryption key.

        Raises:
            Exception: If the profile does not have a signing key.
        """
        if signing_profile.private_key is None:
            raise Exception("Can't make squeak with a contact profile.")
        signing_key_str = signing_profile.private_key.decode()
        signing_key = CSigningKey(signing_key_str)
        block_info = self.bitcoin_client.get_best_block_info()
        block_height = block_info.block_height
        block_hash = block_info.block_hash
        return make_squeak_with_block(
            signing_key,
            content_str,
            block_height,
            block_hash,
            replyto_hash,
        )

    def check_squeak(self, squeak: CSqueak) -> None:
        """Checks if the squeak is valid and has a valid signature.

        Args:
            squeak: The squeak to be validated.

        Returns:
            None

        Raises:
            Exception: If the squeak is not valid.
        """
        check_squeak(squeak)

    def get_block_header(self, squeak: CSqueak) -> CBlockHeader:
        """Checks if the embedded block hash in the squeak is valid for its
        block height and return the associtated block header.

        Args:
            squeak: The squeak to be validated.

        Returns:
            CBlockHeader: the block header associated with the given squeak.

        Raises:
            Exception: If the block hash is not valid.
        """
        block_info = self.bitcoin_client.get_block_info_by_height(
            squeak.nBlockHeight)
        if squeak.hashBlock != block_info.block_hash:
            raise Exception("Block hash incorrect.")
        return block_info.block_header

    def get_decrypted_content(self, squeak: CSqueak, secret_key: bytes) -> str:
        """Checks if the secret key is valid for the given squeak and returns
        the decrypted content.

        Args:
            squeak: The squeak to be validated.
            secret_key: The secret key.

        Returns:
            bytes: the decrypted content

        Raises:
            Exception: If the secret key is not valid.
        """
        return get_decrypted_content(squeak, secret_key)

    def get_best_block_height(self) -> int:
        """Get the current height of the latest block in the blockchain.

        Returns:
            int: the current latest block height.
        """
        block_info = self.bitcoin_client.get_best_block_info()
        return block_info.block_height

    def create_offer(
            self,
            squeak: CSqueak,
            secret_key: bytes,
            peer_address: PeerAddress,
            price_msat: int,
            nonce: bytes = None,
    ) -> SentOffer:
        """Creates an offer to sell a squeak key to another node.

        Args:
            squeak: The squeak to be sold.
            secret_key: The secret key to the squeak.
            peer_address: The address of the buyer.
            price_msat: The price in msats.

        Returns:
            SentOffer: A record of the details of the offer for the seller.
        """
        # Get the squeak hash
        squeak_hash = get_hash(squeak)
        # Generate a new random nonce
        if nonce is None:
            nonce = generate_tweak()
        # Calculate the preimage
        preimage = add_tweak(secret_key, nonce)
        # Create the lightning invoice
        invoice = self.lightning_client.create_invoice(
            preimage, price_msat
        )
        return SentOffer(
            sent_offer_id=None,
            squeak_hash=squeak_hash,
            payment_hash=invoice.r_hash,
            nonce=nonce,
            price_msat=price_msat,
            payment_request=invoice.payment_request,
            invoice_time=invoice.creation_date,
            invoice_expiry=invoice.expiry,
            peer_address=peer_address,
            paid=False,
        )

    def get_lnd_external_address(self) -> Optional[LightningAddressHostPort]:
        """Get the external address of the local lightning node.

        Returns:
            Optional[LightningAddressHostPort]: The host and port of the lnd node if possible.
        """
        info = self.lightning_client.get_info()
        for uri in info.uris:
            pubkey, address = uri.split("@")
            host, port_str = address.split(":")
            port = int(port_str)
            return LightningAddressHostPort(
                host=host,
                port=port,
            )
        return None

    def package_offer(self, sent_offer: SentOffer, lnd_external_address: Optional[LightningAddressHostPort]) -> Offer:
        """Package the offer details into a message that will be sent from
        seller to buyer.

        Args:
            sent_offer: The offer that was already generated by the seller.
            lnd_external_address: The host and port of the lnd node if available.

        Returns:
            SentOffer: A record of the details of the offer for the seller.
        """
        if lnd_external_address is None:
            lnd_external_address = self.get_lnd_external_address()
        return Offer(
            squeak_hash=sent_offer.squeak_hash,
            nonce=sent_offer.nonce,
            payment_request=sent_offer.payment_request,
            host=lnd_external_address.host if lnd_external_address else "",
            port=lnd_external_address.port if lnd_external_address else 0,
        )

    def unpack_offer(
            self,
            squeak: CSqueak,
            offer: Offer,
            peer_address: PeerAddress,
            check_payment_point: bool = False,
    ) -> ReceivedOffer:
        """Get the offer details from the message that the buyer
        receives from the seller.

        Args:
            squeak: The squeak that will be unlocked upon payment.
            offer: The offer details received from the seller.
            peer: The peer that sent the offer.

        Returns:
            ReceivedOffer: A record of the details of the offer for the buyer.
        """
        # Get the squeak hash
        squeak_hash = get_hash(squeak)
        # Check if squeak hash matches squeak_hash in buy_offer.
        if squeak_hash != offer.squeak_hash:
            raise Exception("Squeak hash in offer {!r} does not match squeak hash {!r}.".format(
                offer.squeak_hash, squeak_hash
            ))
        # Decode the payment request
        pay_req = self.lightning_client.decode_pay_req(
            offer.payment_request)
        squeak_payment_point = squeak.paymentPoint
        payment_hash = pay_req.payment_hash
        price_msat = pay_req.num_msat
        destination = pay_req.destination
        invoice_timestamp = pay_req.timestamp
        invoice_expiry = pay_req.expiry
        lightning_address = LightningAddressHostPort(
            host=offer.host or peer_address.host,
            port=offer.port,
        )
        payment_point = pay_req.payment_point
        expected_payment_point = squeak.paymentPoint
        if check_payment_point:
            if payment_point != expected_payment_point:
                raise Exception("Invalid payment point.")
        return ReceivedOffer(
            received_offer_id=None,
            squeak_hash=squeak_hash,
            price_msat=price_msat,
            payment_hash=payment_hash,
            nonce=offer.nonce,
            payment_point=squeak_payment_point,
            invoice_timestamp=invoice_timestamp,
            invoice_expiry=invoice_expiry,
            payment_request=offer.payment_request,
            destination=destination,
            lightning_address=lightning_address,
            peer_address=peer_address,
            paid=False,
        )

    def pay_offer(self, received_offer: ReceivedOffer) -> SentPayment:
        """Pay the offer that the buyer received from the seller.

        Args:
            received_offer: The details of the offer received by the buyer.

        Returns:
            SentPayment: A record of the sent payment.
        """
        # Pay the invoice
        payment = self.lightning_client.pay_invoice(
            received_offer.payment_request)
        preimage = payment.payment_preimage
        if not preimage:
            raise Exception(
                "Payment failed with error: {}".format(payment.payment_error)
            )
        # Calculate the secret key
        nonce = received_offer.nonce
        secret_key = subtract_tweak(preimage, nonce)
        # Check if the secret key is valid for the preimage
        point = get_payment_point_of_secret_key(secret_key)
        valid = point == received_offer.payment_point
        # Save the preimage of the sent payment
        # peer_address = PeerAddress(
        #     network=received_offer.peer_address.network,
        #     host=received_offer.peer_address.host,
        #     port=received_offer.peer_address.port,
        # )
        return SentPayment(
            sent_payment_id=None,
            created_time_ms=None,
            peer_address=received_offer.peer_address,
            squeak_hash=received_offer.squeak_hash,
            payment_hash=received_offer.payment_hash,
            secret_key=secret_key,
            price_msat=received_offer.price_msat,
            node_pubkey=received_offer.destination,
            valid=valid,
        )

    def get_received_payments(
            self,
            latest_settle_index: int,
            get_sent_offer_fn: Callable[[bytes], SentOffer],
    ) -> ReceivedPaymentsStream:
        """Get an iterator of received payments.

        Args:
            latest_settle_index: The latest settle index of the lnd invoice database.
            get_sent_offer_fn: Function that takes a payment hash and returns
                the corresponding SentOffer.

        Returns:
            ReceivedPaymentsStream: An object containing an iterator of received
            payments and a callback function to cancel the iteration.
        """
        # Get the stream of settled invoices.
        invoice_stream = self.lightning_client.subscribe_invoices(
            settle_index=latest_settle_index,
        )

        def cancel_subscription():
            invoice_stream.cancel()

        def get_payment_stream():
            # Yield the received payments.
            try:
                for invoice in invoice_stream.result_stream:
                    if invoice.settled:
                        payment_hash = invoice.r_hash
                        settle_index = invoice.settle_index
                        sent_offer = get_sent_offer_fn(payment_hash)
                        yield ReceivedPayment(
                            received_payment_id=None,
                            created_time_ms=None,
                            squeak_hash=sent_offer.squeak_hash,
                            payment_hash=sent_offer.payment_hash,
                            price_msat=sent_offer.price_msat,
                            settle_index=settle_index,
                            peer_address=sent_offer.peer_address,
                        )
            except grpc.RpcError as e:
                if e.code() != grpc.StatusCode.CANCELLED:
                    raise InvoiceSubscriptionError()

        return ReceivedPaymentsStream(
            cancel_fn=cancel_subscription,
            result_stream=get_payment_stream(),
        )



# MIT License
#
# Copyright (c) 2020 Jonathan Zernik
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
import threading
from concurrent import futures

import grpc

from proto import squeak_admin_pb2_grpc

logger = logging.getLogger(__name__)


class SqueakAdminServerServicer(squeak_admin_pb2_grpc.SqueakAdminServicer):
    """Provides methods that implement functionality of squeak admin server."""

    def __init__(self, host, port, handler):
        self.host = host
        self.port = port
        self.handler = handler
        self.server = None

    def start(self):
        self.server = grpc.server(futures.ThreadPoolExecutor(max_workers=100))
        squeak_admin_pb2_grpc.add_SqueakAdminServicer_to_server(
            self, self.server)
        self.server.add_insecure_port("{}:{}".format(self.host, self.port))
        logger.info("Starting SqueakAdminServerServicer...")
        self.server.start()

    def stop(self):
        if self.server is None:
            return
        self.server.stop(None)
        logger.info("Stopped SqueakAdminServerServicer.")

    def LndGetInfo(self, request, context):
        return self.handler.handle_lnd_get_info(request)

    def LndWalletBalance(self, request, context):
        return self.handler.handle_lnd_wallet_balance(request)

    def LndNewAddress(self, request, context):
        return self.handler.handle_lnd_new_address(request)

    def LndListChannels(self, request, context):
        return self.handler.handle_lnd_list_channels(request)

    def LndPendingChannels(self, request, context):
        return self.handler.handle_lnd_pending_channels(request)

    def LndGetTransactions(self, request, context):
        return self.handler.handle_lnd_get_transactions(request)

    def LndListPeers(self, request, context):
        return self.handler.handle_lnd_list_peers(request)

    def LndConnectPeer(self, request, context):
        return self.handler.handle_lnd_connect_peer(request)

    def LndDisconnectPeer(self, request, context):
        return self.handler.handle_lnd_disconnect_peer(request)

    def LndOpenChannelSync(self, request, context):
        return self.handler.handle_lnd_open_channel_sync(request)

    def LndCloseChannel(self, request, context):
        return self.handler.handle_lnd_close_channel(request)

    def LndSubscribeChannelEvents(self, request, context):
        return self.handler.handle_lnd_subscribe_channel_events(request)

    def LndSendCoins(self, request, context):
        return self.handler.handle_lnd_send_coins(request)

    def CreateSigningProfile(self, request, context):
        return self.handler.handle_create_signing_profile(request)

    def ImportSigningProfile(self, request, context):
        return self.handler.handle_import_signing_profile(request)

    def CreateContactProfile(self, request, context):
        return self.handler.handle_create_contact_profile(request)

    def GetProfiles(self, request, context):
        return self.handler.handle_get_profiles(request)

    def GetSigningProfiles(self, request, context):
        return self.handler.handle_get_signing_profiles(request)

    def GetContactProfiles(self, request, context):
        return self.handler.handle_get_contact_profiles(request)

    def GetSqueakProfile(self, request, context):
        return self.handler.handle_get_squeak_profile(request)

    def GetSqueakProfileByAddress(self, request, context):
        return self.handler.handle_get_squeak_profile_by_address(request)

    def GetSqueakProfileByName(self, request, context):
        return self.handler.handle_get_squeak_profile_by_name(request)

    def SetSqueakProfileFollowing(self, request, context):
        return self.handler.handle_set_squeak_profile_following(request)

    def RenameSqueakProfile(self, request, context):
        return self.handler.handle_rename_squeak_profile(request)

    def DeleteSqueakProfile(self, request, context):
        return self.handler.handle_delete_squeak_profile(request)

    def SetSqueakProfileImage(self, request, context):
        return self.handler.handle_set_squeak_profile_image(request)

    def ClearSqueakProfileImage(self, request, context):
        return self.handler.handle_clear_squeak_profile_image(request)

    def GetSqueakProfilePrivateKey(self, request, context):
        return self.handler.handle_get_squeak_profile_private_key(request)

    def MakeSqueak(self, request, context):
        return self.handler.handle_make_squeak(request)

    def GetSqueakDisplay(self, request, context):
        return self.handler.handle_get_squeak_display_entry(request)

    def GetTimelineSqueakDisplays(self, request, context):
        return self.handler.handle_get_timeline_squeak_display_entries(request)

    def GetAddressSqueakDisplays(self, request, context):
        return self.handler.handle_get_squeak_display_entries_for_address(request)

    def GetSearchSqueakDisplays(self, request, context):
        return self.handler.handle_get_squeak_display_entries_for_text_search(request)

    def GetAncestorSqueakDisplays(self, request, context):
        return self.handler.handle_get_ancestor_squeak_display_entries(request)

    def GetReplySqueakDisplays(self, request, context):
        return self.handler.handle_get_reply_squeak_display_entries(request)

    def DeleteSqueak(self, request, context):
        return self.handler.handle_delete_squeak(request)

    def CreatePeer(self, request, context):
        return self.handler.handle_create_peer(request)

    def GetPeer(self, request, context):
        return self.handler.handle_get_squeak_peer(request)

    def GetPeerByAddress(self, request, context):
        return self.handler.handle_get_squeak_peer_by_address(request)

    def GetPeers(self, request, context):
        return self.handler.handle_get_squeak_peers(request)

    def SetPeerAutoconnect(self, request, context):
        return self.handler.handle_set_squeak_peer_autoconnect(request)

    def SetPeerShareForFree(self, request, context):
        return self.handler.handle_set_squeak_peer_share_for_free(request)

    def RenamePeer(self, request, context):
        return self.handler.handle_rename_squeak_peer(request)

    def DeletePeer(self, request, context):
        return self.handler.handle_delete_squeak_peer(request)

    def GetBuyOffers(self, request, context):
        return self.handler.handle_get_buy_offers(request)

    def GetBuyOffer(self, request, context):
        return self.handler.handle_get_buy_offer(request)

    def DownloadSqueaks(self, request, context):
        return self.handler.handle_download_squeaks(request)

    def DownloadSqueak(self, request, context):
        return self.handler.handle_download_squeak(request)

    def DownloadOffers(self, request, context):
        return self.handler.handle_download_offers(request)

    def DownloadReplies(self, request, context):
        return self.handler.handle_download_replies(request)

    def DownloadAddressSqueaks(self, request, context):
        return self.handler.handle_download_address_squeaks(request)

    def PayOffer(self, request, context):
        return self.handler.handle_pay_offer(request)

    def GetSentPayments(self, request, context):
        return self.handler.handle_get_sent_payments(request)

    def GetSentPayment(self, request, context):
        return self.handler.handle_get_sent_payment(request)

    def GetSqueakDetails(self, request, context):
        return self.handler.handle_get_squeak_details(request)

    def GetSentOffers(self, request, context):
        return self.handler.handle_get_sent_offers(request)

    def GetReceivedPayments(self, request, context):
        return self.handler.handle_get_received_payments(request)

    def SubscribeReceivedPayments(self, request, context):
        stopped = threading.Event()

        def on_rpc_done():
            logger.info("Stopping SubscribeReceivedPayments.")
            stopped.set()
        context.add_callback(on_rpc_done)
        return self.handler.handle_subscribe_received_payments(
            request,
            stopped,
        )

    def GetNetwork(self, request, context):
        return self.handler.handle_get_network(request)

    def GetPaymentSummary(self, request, context):
        return self.handler.handle_get_payment_summary(request)

    def ReprocessReceivedPayments(self, request, context):
        return self.handler.handle_reprocess_received_payments(request)

    def LikeSqueak(self, request, context):
        return self.handler.handle_like_squeak(request)

    def UnlikeSqueak(self, request, context):
        return self.handler.handle_unlike_squeak(request)

    def GetLikedSqueakDisplays(self, request, context):
        return self.handler.handle_get_liked_squeak_display_entries(request)

    def ConnectPeer(self, request, context):
        return self.handler.handle_connect_peer(request)

    def GetConnectedPeers(self, request, context):
        return self.handler.handle_get_connected_peers(request)

    def GetConnectedPeer(self, request, context):
        return self.handler.handle_get_connected_peer(request)

    def DisconnectPeer(self, request, context):
        return self.handler.handle_disconnect_peer(request)

    def SubscribeConnectedPeers(self, request, context):
        stopped = threading.Event()

        def on_rpc_done():
            logger.info("Stopping SubscribeConnectedPeers.")
            stopped.set()
        context.add_callback(on_rpc_done)
        return self.handler.handle_subscribe_connected_peers(
            request,
            stopped,
        )

    def SubscribeConnectedPeer(self, request, context):
        stopped = threading.Event()

        def on_rpc_done():
            logger.info("Stopping SubscribeConnectedPeer.")
            stopped.set()
        context.add_callback(on_rpc_done)
        return self.handler.handle_subscribe_connected_peer(
            request,
            stopped,
        )

    def SubscribeBuyOffers(self, request, context):
        stopped = threading.Event()

        def on_rpc_done():
            logger.info("Stopping SubscribeBuyOffers.")
            stopped.set()
        context.add_callback(on_rpc_done)
        return self.handler.handle_subscribe_buy_offers(
            request,
            stopped,
        )

    def SubscribeSqueakDisplay(self, request, context):
        stopped = threading.Event()

        def on_rpc_done():
            logger.info("Stopping SubscribeSqueakDisplay.")
            stopped.set()
        context.add_callback(on_rpc_done)
        return self.handler.handle_subscribe_squeak_display(
            request,
            stopped,
        )

    def SubscribeReplySqueakDisplays(self, request, context):
        stopped = threading.Event()

        def on_rpc_done():
            logger.info("Stopping SubscribeReplySqueakDisplays.")
            stopped.set()
        context.add_callback(on_rpc_done)
        return self.handler.handle_subscribe_reply_squeak_displays(
            request,
            stopped,
        )

    def SubscribeAddressSqueakDisplays(self, request, context):
        stopped = threading.Event()

        def on_rpc_done():
            logger.info("Stopping SubscribeAddressSqueakDisplaysRequest.")
            stopped.set()
        context.add_callback(on_rpc_done)
        return self.handler.handle_subscribe_address_squeak_displays(
            request,
            stopped,
        )

    def SubscribeAncestorSqueakDisplays(self, request, context):
        stopped = threading.Event()

        def on_rpc_done():
            logger.info("Stopping SubscribeAncestorSqueakDisplays.")
            stopped.set()
        context.add_callback(on_rpc_done)
        return self.handler.handle_subscribe_ancestor_squeak_displays(
            request,
            stopped,
        )

    def SubscribeSqueakDisplays(self, request, context):
        stopped = threading.Event()

        def on_rpc_done():
            logger.info("Stopping SubscribeSqueakDisplays.")
            stopped.set()
        context.add_callback(on_rpc_done)
        return self.handler.handle_subscribe_squeak_displays(
            request,
            stopped,
        )

    def SubscribeTimelineSqueakDisplays(self, request, context):
        stopped = threading.Event()

        def on_rpc_done():
            logger.info("Stopping SubscribeTimelineSqueakDisplays.")
            stopped.set()
        context.add_callback(on_rpc_done)
        return self.handler.handle_subscribe_timeline_squeak_displays(
            request,
            stopped,
        )

    def GetExternalAddress(self, request, context):
        return self.handler.handle_get_external_address(request)

    def GetDefaultPeerPort(self, request, context):
        return self.handler.handle_get_default_peer_port(request)

    def SetSellPrice(self, request, context):
        return self.handler.handle_set_sell_price(request)

    def ClearSellPrice(self, request, context):
        return self.handler.handle_clear_sell_price(request)

    def GetSellPrice(self, request, context):
        return self.handler.handle_get_sell_price(request)

    def SetTwitterBearerToken(self, request, context):
        return self.handler.handle_set_twitter_bearer_token(request)

    def GetTwitterBearerToken(self, request, context):
        return self.handler.handle_get_twitter_bearer_token(request)

    def AddTwitterAccount(self, request, context):
        return self.handler.handle_add_twitter_account(request)

    def GetTwitterAccounts(self, request, context):
        return self.handler.handle_get_twitter_accounts(request)

    def DeleteTwitterAccount(self, request, context):
        return self.handler.handle_delete_twitter_account(request)



# MIT License
#
# Copyright (c) 2020 Jonathan Zernik
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import hashlib
import os
import random
import uuid

from bitcoin.core import CBlockHeader
from squeak.core.signing import CSigningKey
from squeak.core.signing import CSqueakAddress

from squeaknode.core.peer_address import Network
from squeaknode.core.peer_address import PeerAddress
from squeaknode.core.peers import create_saved_peer
from squeaknode.core.profiles import create_contact_profile
from squeaknode.core.profiles import create_signing_profile
from squeaknode.core.received_payment import ReceivedPayment
from squeaknode.core.sent_payment import SentPayment
from squeaknode.core.squeaks import HASH_LENGTH
from squeaknode.core.squeaks import make_squeak_with_block


def gen_signing_key():
    return CSigningKey.generate()


def gen_random_hash():
    return os.urandom(HASH_LENGTH)


def sha256(data):
    return hashlib.sha256(data).digest()


def address_from_signing_key(signing_key):
    verifying_key = signing_key.get_verifying_key()
    return CSqueakAddress.from_verifying_key(verifying_key)


def gen_address():
    signing_key = gen_signing_key()
    return address_from_signing_key(signing_key)


def gen_squeak_addresses(n):
    return [gen_address() for i in range(n)]


def gen_squeak(signing_key, block_height, replyto_hash=None):
    random_content = "random_content_{}".format(uuid.uuid1())
    random_hash = gen_random_hash()
    squeak, secret_key = make_squeak_with_block(
        signing_key,
        random_content,
        block_height,
        random_hash,
        replyto_hash=replyto_hash,
    )
    return squeak


def gen_block_header(block_height):
    return CBlockHeader(
        nTime=block_height * 10,  # So that block times are increasing.
    )


def gen_squeak_with_block_header(signing_key, block_height, replyto_hash=None):
    """ Return a tuple with a CSqueak and a CBlockHeader.
    """
    squeak = gen_squeak(
        signing_key=signing_key,
        block_height=block_height,
        replyto_hash=replyto_hash,
    )
    block_info = gen_block_header(
        block_height=block_height,
    )
    return squeak, block_info


def gen_signing_profile(profile_name, signing_key):
    return create_signing_profile(
        profile_name,
        signing_key,
    )


def gen_contact_profile(profile_name, address):
    return create_contact_profile(
        profile_name,
        address,
    )


def gen_squeak_peer(peer_name):
    host = "random_host_{}".format(uuid.uuid1())
    port = random.randint(1, 10000)
    peer_address = PeerAddress(
        network=Network.IPV4,
        host=host,
        port=port,
    )
    return create_saved_peer(
        peer_name,
        peer_address,
    )


def gen_sent_payment(peer_address, squeak_hash, secret_key, price_msat, seller_pubkey):
    payment_hash = gen_random_hash()
    return SentPayment(
        sent_payment_id=None,
        created_time_ms=None,
        peer_address=peer_address,
        squeak_hash=squeak_hash,
        payment_hash=payment_hash,
        secret_key=secret_key,
        price_msat=price_msat,
        node_pubkey=seller_pubkey,
        valid=True,
    )


def gen_received_payment(peer_address, squeak_hash, price_msat, settle_index):
    payment_hash = gen_random_hash()
    return ReceivedPayment(
        received_payment_id=None,
        created_time_ms=None,
        squeak_hash=squeak_hash,
        payment_hash=payment_hash,
        price_msat=price_msat,
        settle_index=settle_index,
        peer_address=peer_address,
    )



# MIT License
#
# Copyright (c) 2020 Jonathan Zernik
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from squeak.messages import msg_offer
from squeak.messages import msg_secretkey
from squeak.messages import MsgSerializable

from squeaknode.core.offer import Offer


class SecretKeyReply:

    def get_msg(self) -> MsgSerializable:
        """Return the message to reply to the other peer.
        """


class FreeSecretKeyReply(SecretKeyReply):

    def __init__(self, squeak_hash: bytes, secret_key: bytes):
        self.squeak_hash = squeak_hash
        self.secret_key = secret_key

    def get_msg(self) -> MsgSerializable:
        return msg_secretkey(
            hashSqk=self.squeak_hash,
            secretKey=self.secret_key,
        )


class OfferReply(SecretKeyReply):

    def __init__(self, squeak_hash: bytes, offer: Offer):
        self.squeak_hash = squeak_hash
        self.offer = offer

    def get_msg(self) -> MsgSerializable:
        return msg_offer(
            hashSqk=self.squeak_hash,
            nonce=self.offer.nonce,
            strPaymentInfo=self.offer.payment_request.encode(
                'utf-8'),
            host=self.offer.host.encode('utf-8'),
            port=self.offer.port,
        )



# MIT License
#
# Copyright (c) 2020 Jonathan Zernik
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
import threading
import time
import uuid
from abc import ABC
from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import TimeoutError
from typing import Dict
from typing import Optional

from squeak.messages import msg_getdata
from squeak.messages import msg_getsqueaks
from squeak.messages import MsgSerializable
from squeak.net import CInterested
from squeak.net import CInv
from squeak.net import CSqueakLocator

from squeaknode.core.download_result import DownloadResult
from squeaknode.node.downloaded_object import DownloadedObject

logger = logging.getLogger(__name__)


DOWNLOAD_TIMEOUT_S = 10


class ActiveDownload(ABC):

    def __init__(self, limit: int):
        self.limit = limit
        self.count = 0
        self._lock = threading.Lock()
        self.stopped = threading.Event()
        self.num_peers = 0
        self.start_time_ms: Optional[int] = None

    @abstractmethod
    def is_interested(self, downloaded_object: DownloadedObject) -> bool:
        """Return True if the given squeak matches the download interest."""

    @abstractmethod
    def get_download_msg(self) -> MsgSerializable:
        """Get the message to send to peers to get download response."""

    def initiate_download(self, broadcast_fn) -> None:
        self.start_time_ms = int(time.time() * 1000)
        msg = self.get_download_msg()
        self.num_peers = broadcast_fn(msg)
        if self.num_peers == 0:
            self.mark_complete()

    def increment(self) -> None:
        with self._lock:
            self.count += 1
            if self.count >= self.limit:
                self.mark_complete()

    def mark_complete(self):
        self.stopped.set()

    def cancel(self):
        self.stopped.set()

    def get_elapsed_time_ms(self):
        if self.start_time_ms is None:
            return 0
        end_time_ms = int(time.time() * 1000)
        return end_time_ms - self.start_time_ms

    def wait_for_complete(self, timeout_s: int) -> None:
        self.stopped.wait(timeout=timeout_s)

    def get_result(self) -> DownloadResult:
        return DownloadResult(
            number_downloaded=self.count,
            number_requested=self.limit,
            elapsed_time_ms=self.get_elapsed_time_ms(),
            number_peers=self.num_peers,
        )


class InterestDownload(ActiveDownload):

    def __init__(self, limit: int, interest: CInterested):
        self.interest = interest
        super().__init__(limit)

    def is_interested(self, downloaded_object: DownloadedObject) -> bool:
        return downloaded_object.matches_requested_squeak_range(self.interest)

    def get_download_msg(self) -> MsgSerializable:
        locator = CSqueakLocator(
            vInterested=[self.interest],
        )
        return msg_getsqueaks(
            locator=locator,
        )


class HashDownload(ActiveDownload):

    def __init__(self, squeak_hash: bytes):
        self.squeak_hash = squeak_hash
        super().__init__(1)

    def is_interested(self, downloaded_object: DownloadedObject) -> bool:
        return downloaded_object.matches_requested_squeak_hash(self.squeak_hash)

    def get_download_msg(self) -> MsgSerializable:
        invs = [
            CInv(type=1, hash=self.squeak_hash)
        ]
        return msg_getdata(
            inv=invs,
        )


class OffersDownload(ActiveDownload):

    def __init__(self, limit: int, squeak_hash: bytes):
        self.squeak_hash = squeak_hash
        super().__init__(limit)

    def is_interested(self, downloaded_object: DownloadedObject) -> bool:
        return downloaded_object.matches_requested_offer_hash(self.squeak_hash)

    def get_download_msg(self) -> MsgSerializable:
        invs = [
            CInv(type=2, hash=self.squeak_hash)
        ]
        return msg_getdata(inv=invs)


class ActiveDownloadManager:

    def __init__(self):
        self.downloads: Dict[str, ActiveDownload] = dict()
        self.executor = None
        self.broadcast_fn = None

    def start(self, broadcast_fn):
        logger.info("Starting Download Manager...")
        self.broadcast_fn = broadcast_fn
        self.executor = ThreadPoolExecutor(max_workers=10)

    def stop(self):
        for download in self.downloads.values():
            download.cancel()
        logger.info("Stopping Download Manager...")
        self.executor.shutdown(wait=True)
        logger.info("Stopped Download Manager.")

    def lookup_counter(self, downloaded_object: DownloadedObject) -> Optional[ActiveDownload]:
        for name, interest in self.downloads.items():
            if interest.is_interested(downloaded_object):
                return interest
        return None

    def run_download(self, download: ActiveDownload) -> DownloadResult:
        name_key = "download_key_{}".format(uuid.uuid1())
        self.downloads[name_key] = download
        future = self.executor.submit(self.download_task, download)
        try:
            return future.result()
        except TimeoutError:
            return download.get_result()
        finally:
            del self.downloads[name_key]

    def download_task(self, download: ActiveDownload) -> DownloadResult:
        download.initiate_download(self.broadcast_fn)
        download.wait_for_complete(DOWNLOAD_TIMEOUT_S)
        return download.get_result()

    def download_interest(self, limit: int, interest: CInterested) -> DownloadResult:
        download = InterestDownload(limit, interest)
        return self.run_download(download)

    def download_hash(self, squeak_hash: bytes) -> DownloadResult:
        download = HashDownload(squeak_hash)
        return self.run_download(download)

    def download_offers(self, limit: int, squeak_hash: bytes) -> DownloadResult:
        download = OffersDownload(limit, squeak_hash)
        return self.run_download(download)



# MIT License
#
# Copyright (c) 2020 Jonathan Zernik
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
import time
from contextlib import contextmanager
from typing import Iterator
from typing import List
from typing import Optional

import sqlalchemy
from bitcoin.core import CBlockHeader
from sqlalchemy import func
from sqlalchemy import literal
from sqlalchemy import not_
from sqlalchemy import or_
from sqlalchemy.sql import select
from sqlalchemy.sql import tuple_
from squeak.core import CSqueak

from squeaknode.core.lightning_address import LightningAddressHostPort
from squeaknode.core.peer_address import Network
from squeaknode.core.peer_address import PeerAddress
from squeaknode.core.received_offer import ReceivedOffer
from squeaknode.core.received_payment import ReceivedPayment
from squeaknode.core.received_payment_summary import ReceivedPaymentSummary
from squeaknode.core.sent_offer import SentOffer
from squeaknode.core.sent_payment import SentPayment
from squeaknode.core.sent_payment_summary import SentPaymentSummary
from squeaknode.core.squeak_entry import SqueakEntry
from squeaknode.core.squeak_peer import SqueakPeer
from squeaknode.core.squeak_profile import SqueakProfile
from squeaknode.core.squeaks import get_hash
from squeaknode.core.twitter_account import TwitterAccount
from squeaknode.core.twitter_account_entry import TwitterAccountEntry
from squeaknode.core.user_config import UserConfig
from squeaknode.db.exception import SqueakDatabaseError
from squeaknode.db.migrations import run_migrations
from squeaknode.db.models import Models


MAX_INT = 999999999999
MAX_HASH = b'\xff' * 32
INIT_NUM_RETRIES = 10
INIT_RETRY_INTERVAL_S = 1


logger = logging.getLogger(__name__)


class SqueakDb:
    def __init__(self, engine, schema=None):
        self.engine = engine
        self.schema = schema
        self.models = Models(schema=schema)

    @contextmanager
    def get_connection(self):
        with self.engine.connect() as connection:
            yield connection

    def init(self):
        """ Create the tables and indices in the database. """
        logger.debug("SqlAlchemy version: {}".format(sqlalchemy.__version__))
        run_migrations(self.engine)

    def init_with_retries(
            self,
            num_retries=INIT_NUM_RETRIES,
            retry_interval_s=INIT_RETRY_INTERVAL_S,
    ):
        """ Try repeatedly to init the database.

        Raises exception if db init fails more than `num_retries` times.
        """
        n = 0
        while True:
            try:
                self.init()
                return
            except Exception:
                logger.exception("Failed to initialize database.")
                n += 1
                if n >= num_retries:
                    raise SqueakDatabaseError("Failed to initialize database.")
                time.sleep(retry_interval_s)

    @property
    def squeaks(self):
        return self.models.squeaks

    @property
    def profiles(self):
        return self.models.profiles

    @property
    def peers(self):
        return self.models.peers

    @property
    def received_offers(self):
        return self.models.received_offers

    @property
    def sent_payments(self):
        return self.models.sent_payments

    @property
    def received_payments(self):
        return self.models.received_payments

    @property
    def sent_offers(self):
        return self.models.sent_offers

    @property
    def configs(self):
        return self.models.configs

    @property
    def twitter_accounts(self):
        return self.models.twitter_accounts

    @property
    def squeak_has_secret_key(self):
        return self.squeaks.c.secret_key != None  # noqa: E711

    @property
    def squeak_is_liked(self):
        return self.squeaks.c.liked_time_ms != None  # noqa: E711

    def squeak_is_older_than_retention(self, interval_s):
        return self.timestamp_now_ms > \
            self.squeaks.c.created_time_ms + interval_s * 1000

    @property
    def profile_has_private_key(self):
        return self.profiles.c.private_key != None  # noqa: E711

    @property
    def profile_is_following(self):
        return self.profiles.c.following == True  # noqa: E711

    @property
    def timestamp_now_ms(self):
        return int(time.time() * 1000)

    @property
    def received_offer_invoice_is_expired(self):
        expire_time = (
            self.received_offers.c.invoice_timestamp
            + self.received_offers.c.invoice_expiry
        )
        return self.timestamp_now_ms / 1000 >= expire_time

    def received_offer_is_out_of_retention(self, interval_s):
        expire_time_s = (
            self.received_offers.c.created_time_ms / 1000
            + interval_s
        )
        return self.timestamp_now_ms / 1000 >= expire_time_s

    @property
    def received_offer_is_paid(self):
        return self.received_offers.c.paid == True  # noqa: E711

    def sent_offer_out_of_retention(self, interval_s):
        expire_time = (
            self.sent_offers.c.invoice_timestamp
            + self.sent_offers.c.invoice_expiry
            + interval_s
        )
        return self.timestamp_now_ms / 1000 >= expire_time

    @property
    def sent_offer_is_paid(self):
        return self.sent_offers.c.paid == True  # noqa: E711

    @property
    def sent_offer_is_expired(self):
        expire_time = (
            self.sent_offers.c.invoice_timestamp
            + self.sent_offers.c.invoice_expiry
        )
        return self.timestamp_now_ms / 1000 >= expire_time

    def insert_squeak(self, squeak: CSqueak, block_header: CBlockHeader) -> Optional[bytes]:
        """ Insert a new squeak.

        Return the hash (bytes) of the inserted squeak.
        Return None if squeak already exists.
        """
        ins = self.squeaks.insert().values(
            created_time_ms=self.timestamp_now_ms,
            hash=get_hash(squeak),
            squeak=squeak.serialize(),
            hash_reply_sqk=squeak.hashReplySqk if squeak.is_reply else None,
            hash_block=squeak.hashBlock,
            n_block_height=squeak.nBlockHeight,
            n_time=squeak.nTime,
            author_address=str(squeak.GetAddress()),
            secret_key=None,
            block_time=block_header.nTime,
        )
        with self.get_connection() as connection:
            try:
                res = connection.execute(ins)
                squeak_hash = res.inserted_primary_key[0]
                return squeak_hash
            except sqlalchemy.exc.IntegrityError:
                logger.debug("Failed to insert squeak.", exc_info=True)
                return None

    def get_squeak(self, squeak_hash: bytes) -> Optional[CSqueak]:
        """ Get a squeak. """
        s = select([self.squeaks]).where(
            self.squeaks.c.hash == squeak_hash)
        with self.get_connection() as connection:
            result = connection.execute(s)
            row = result.fetchone()
            if row is None:
                return None
            return self._parse_squeak(row)

    def get_squeak_secret_key(self, squeak_hash: bytes) -> Optional[bytes]:
        """ Get a squeak secret key. """
        s = select([self.squeaks]).where(
            self.squeaks.c.hash == squeak_hash)
        with self.get_connection() as connection:
            result = connection.execute(s)
            row = result.fetchone()
            if row is None:
                return None
            return row["secret_key"]

    def get_squeak_entry(self, squeak_hash: bytes) -> Optional[SqueakEntry]:
        """ Get a squeak with the author profile. """
        s = (
            select([self.squeaks, self.profiles])
            .select_from(
                self.squeaks.outerjoin(
                    self.profiles,
                    self.profiles.c.address == self.squeaks.c.author_address,
                )
            )
            .where(self.squeaks.c.hash == squeak_hash)
        )
        with self.get_connection() as connection:
            result = connection.execute(s)
            row = result.fetchone()
            if row is None:
                return None
            return self._parse_squeak_entry(row)

    def get_timeline_squeak_entries(
            self,
            limit: int,
            last_entry: Optional[SqueakEntry],
    ) -> List[SqueakEntry]:
        """ Get all followed squeaks. """
        last_block_height = last_entry.block_height if last_entry else MAX_INT
        last_squeak_time = last_entry.squeak_time if last_entry else MAX_INT
        last_squeak_hash = last_entry.squeak_hash if last_entry else MAX_HASH
        logger.info("""Timeline db query with
        limit: {}
        block_height: {}
        squeak_time: {}
        squeak_hash: {}
        """.format(
            limit,
            last_block_height,
            last_squeak_time,
            last_squeak_hash.hex(),
        ))
        s = (
            select([self.squeaks, self.profiles])
            .select_from(
                self.squeaks.outerjoin(
                    self.profiles,
                    self.profiles.c.address == self.squeaks.c.author_address,
                )
            )
            .where(self.profile_is_following)
            .where(
                tuple_(
                    self.squeaks.c.n_block_height,
                    self.squeaks.c.n_time,
                    self.squeaks.c.hash,
                ) < tuple_(
                    last_block_height,
                    last_squeak_time,
                    last_squeak_hash,
                )
            )
            .order_by(
                self.squeaks.c.n_block_height.desc(),
                self.squeaks.c.n_time.desc(),
                self.squeaks.c.hash.desc(),
            )
            .limit(limit)
        )
        with self.get_connection() as connection:
            result = connection.execute(s)
            rows = result.fetchall()
            return [self._parse_squeak_entry(row) for row in rows]

    def get_liked_squeak_entries(
            self,
            limit: int,
            last_entry: Optional[SqueakEntry],
    ) -> List[SqueakEntry]:
        """ Get liked squeaks. """
        last_liked_time_ms = last_entry.liked_time_ms if last_entry else self.timestamp_now_ms
        last_squeak_hash = last_entry.squeak_hash if last_entry else MAX_HASH
        logger.info("""Liked squeaks db query with
        limit: {}
        last_liked_time_ms: {}
        last_squeak_hash: {}
        """.format(
            limit,
            last_liked_time_ms,
            last_squeak_hash.hex(),
        ))
        s = (
            select([self.squeaks, self.profiles])
            .select_from(
                self.squeaks.outerjoin(
                    self.profiles,
                    self.profiles.c.address == self.squeaks.c.author_address,
                )
            )
            .where(
                self.squeak_is_liked,
            )
            .where(
                tuple_(
                    self.squeaks.c.liked_time_ms,
                    self.squeaks.c.hash,
                ) < tuple_(
                    last_liked_time_ms,
                    last_squeak_hash,
                )
            )
            .order_by(
                self.squeaks.c.liked_time_ms.desc(),
                self.squeaks.c.hash.desc(),
            )
            .limit(limit)
        )
        with self.get_connection() as connection:
            result = connection.execute(s)
            rows = result.fetchall()
            return [self._parse_squeak_entry(row) for row in rows]

    def get_squeak_entries_for_address(
            self,
            address: str,
            limit: int,
            last_entry: Optional[SqueakEntry],
    ) -> List[SqueakEntry]:
        """ Get a squeak. """
        last_block_height = last_entry.block_height if last_entry else MAX_INT
        last_squeak_time = last_entry.squeak_time if last_entry else MAX_INT
        last_squeak_hash = last_entry.squeak_hash if last_entry else MAX_HASH
        logger.info("""Address db query with
        address: {}
        limit: {}
        block_height: {}
        squeak_time: {}
        squeak_hash: {}
        """.format(
            address,
            limit,
            last_block_height,
            last_squeak_time,
            last_squeak_hash.hex(),
        ))
        s = (
            select([self.squeaks, self.profiles])
            .select_from(
                self.squeaks.outerjoin(
                    self.profiles,
                    self.profiles.c.address == self.squeaks.c.author_address,
                )
            )
            .where(self.squeaks.c.author_address == address)
            .where(
                tuple_(
                    self.squeaks.c.n_block_height,
                    self.squeaks.c.n_time,
                    self.squeaks.c.hash,
                ) < tuple_(
                    last_block_height,
                    last_squeak_time,
                    last_squeak_hash,
                )
            )
            .order_by(
                self.squeaks.c.n_block_height.desc(),
                self.squeaks.c.n_time.desc(),
                self.squeaks.c.hash.desc(),
            )
            .limit(limit)
        )
        with self.get_connection() as connection:
            result = connection.execute(s)
            rows = result.fetchall()
            return [self._parse_squeak_entry(row) for row in rows]

    def get_squeak_entries_for_text_search(
            self,
            search_text: str,
            limit: int,
            last_entry: Optional[SqueakEntry],
    ) -> List[SqueakEntry]:
        """ Get a squeak. """
        last_block_height = last_entry.block_height if last_entry else MAX_INT
        last_squeak_time = last_entry.squeak_time if last_entry else MAX_INT
        last_squeak_hash = last_entry.squeak_hash if last_entry else MAX_HASH
        logger.info("""Search db query with
        search_text: {}
        limit: {}
        block_height: {}
        squeak_time: {}
        squeak_hash: {}
        """.format(
            search_text,
            limit,
            last_block_height,
            last_squeak_time,
            last_squeak_hash.hex(),
        ))
        s = (
            select([self.squeaks, self.profiles])
            .select_from(
                self.squeaks.outerjoin(
                    self.profiles,
                    self.profiles.c.address == self.squeaks.c.author_address,
                )
            )
            .where(self.squeaks.c.content.ilike(f'%{search_text}%'))
            .where(
                tuple_(
                    self.squeaks.c.n_block_height,
                    self.squeaks.c.n_time,
                    self.squeaks.c.hash,
                ) < tuple_(
                    last_block_height,
                    last_squeak_time,
                    last_squeak_hash,
                )
            )
            .order_by(
                self.squeaks.c.n_block_height.desc(),
                self.squeaks.c.n_time.desc(),
                self.squeaks.c.hash.desc(),
            )
            .limit(limit)
        )
        with self.get_connection() as connection:
            result = connection.execute(s)
            rows = result.fetchall()
            return [self._parse_squeak_entry(row) for row in rows]

    def get_thread_ancestor_squeak_entries(self, squeak_hash: bytes) -> List[SqueakEntry]:
        """ Get all reply ancestors of squeak hash. """
        ancestors = (
            select(
                [
                    self.squeaks.c.hash.label("hash"),
                    literal(0).label("depth"),
                ]
            )
            .where(self.squeaks.c.hash == squeak_hash)
            .cte(recursive=True)
        )

        ancestors_alias = ancestors.alias()
        squeaks_alias = self.squeaks.alias()

        ancestors = ancestors.union_all(
            select(
                [
                    squeaks_alias.c.hash_reply_sqk.label("hash"),
                    (ancestors_alias.c.depth + 1).label("depth"),
                ]
            ).where(squeaks_alias.c.hash == ancestors_alias.c.hash)
        )

        s = (
            select([self.squeaks, self.profiles])
            .select_from(
                self.squeaks.join(
                    ancestors,
                    ancestors.c.hash == self.squeaks.c.hash,
                ).outerjoin(
                    self.profiles,
                    self.profiles.c.address == self.squeaks.c.author_address,
                )
            )
            .order_by(
                ancestors.c.depth.desc(),
            )
        )

        with self.get_connection() as connection:
            result = connection.execute(s)
            rows = result.fetchall()
            return [self._parse_squeak_entry(row) for row in rows]

        # sql = """
        # WITH RECURSIVE is_thread_ancestor(hash, depth) AS (
        #     VALUES(%s, 1)
        #   UNION\n
        #     SELECT squeak.hash_reply_sqk AS hash, is_thread_ancestor.depth + 1 AS depth FROM squeak, is_thread_ancestor
        #     WHERE squeak.hash=is_thread_ancestor.hash
        #   )
        #   SELECT * FROM squeak
        #   JOIN is_thread_ancestor
        #     ON squeak.hash=is_thread_ancestor.hash
        #   LEFT JOIN profile
        #     ON squeak.author_address=profile.address
        #   WHERE squeak.block_header IS NOT NULL
        #   ORDER BY is_thread_ancestor.depth DESC;
        # """
        # with self.get_cursor() as curs:
        #     curs.execute(sql, (squeak_hash_str,))
        #     rows = curs.fetchall()
        #     return [self._parse_squeak_entry_with_profile(row) for row in rows]

    def get_thread_reply_squeak_entries(
            self,
            squeak_hash: bytes,
            limit: int,
            last_entry: Optional[SqueakEntry],
    ) -> List[SqueakEntry]:
        """ Get all replies for a squeak hash. """
        last_block_height = last_entry.block_height if last_entry else MAX_INT
        last_squeak_time = last_entry.squeak_time if last_entry else MAX_INT
        last_squeak_hash = last_entry.squeak_hash if last_entry else MAX_HASH
        logger.info("""Timeline db query with
        limit: {}
        block_height: {}
        squeak_time: {}
        squeak_hash: {}
        """.format(
            limit,
            last_block_height,
            last_squeak_time,
            last_squeak_hash.hex(),
        ))
        s = (
            select([self.squeaks, self.profiles])
            .select_from(
                self.squeaks.outerjoin(
                    self.profiles,
                    self.profiles.c.address == self.squeaks.c.author_address,
                )
            )
            .where(self.squeaks.c.hash_reply_sqk == squeak_hash)
            .where(
                tuple_(
                    self.squeaks.c.n_block_height,
                    self.squeaks.c.n_time,
                    self.squeaks.c.hash,
                ) < tuple_(
                    last_block_height,
                    last_squeak_time,
                    last_squeak_hash,
                )
            )
            .order_by(
                self.squeaks.c.n_block_height.desc(),
                self.squeaks.c.n_time.desc(),
                self.squeaks.c.hash.desc(),
            )
            .limit(limit)
        )
        with self.get_connection() as connection:
            result = connection.execute(s)
            rows = result.fetchall()
            return [self._parse_squeak_entry(row) for row in rows]

    def lookup_squeaks(
        self,
        addresses: List[str],
        min_block: Optional[int],
        max_block: Optional[int],
        reply_to_hash: Optional[bytes],
        include_locked: bool = False,
    ) -> List[bytes]:
        """ Lookup squeaks. """
        logger.debug("""Running lookup query with
        addresses: {},
        min_block: {}
        max_block: {}
        reply_to_hash: {!r}
        include_locked: {}""".format(
            addresses,
            min_block,
            max_block,
            reply_to_hash,
            include_locked,
        ))

        s = select([self.squeaks.c.hash])
        if addresses:
            s = s.where(self.squeaks.c.author_address.in_(addresses))
        if min_block:
            s = s.where(self.squeaks.c.n_block_height >= min_block)
        if max_block:
            s = s.where(self.squeaks.c.n_block_height <= max_block)
        if reply_to_hash:
            s = s.where(self.squeaks.c.hash_reply_sqk == reply_to_hash)
        if not include_locked:
            s = s.where(self.squeak_has_secret_key)
        s = s.order_by(
            self.squeaks.c.n_block_height.desc(),
            self.squeaks.c.n_time.desc(),
            self.squeaks.c.hash.desc(),
        )

        logger.debug("Query s: {}".format(s))

        with self.get_connection() as connection:
            # logger.debug("Mogrify lookup_squeaks: {}.".format(
            #     connection.connection.cursor().mogrify(s),
            # ))
            result = connection.execute(s)
            rows = result.fetchall()
            hashes = [(row["hash"]) for row in rows]
            return hashes

    def get_number_of_squeaks(self) -> int:
        """ Get total number of squeaks. """
        s = (
            select([
                func.count().label("num_squeaks"),
            ])
            .select_from(self.squeaks)
        )
        with self.get_connection() as connection:
            result = connection.execute(s)
            row = result.fetchone()
            num_squeaks = row["num_squeaks"]
            return num_squeaks

    def number_of_squeaks_with_address_in_block_range(
        self,
        address: str,
        min_block: int,
        max_block: int,
    ) -> int:
        """ Get number of squeaks with address in block range. """
        s = (
            select([
                func.count().label("num_squeaks"),
            ])
            .select_from(self.squeaks)
            .where(self.squeaks.c.author_address == address)
            .where(self.squeaks.c.n_block_height >= min_block)
            .where(self.squeaks.c.n_block_height <= max_block)
        )
        with self.get_connection() as connection:
            result = connection.execute(s)
            row = result.fetchone()
            num_squeaks = row["num_squeaks"]
            return num_squeaks

    def get_old_squeaks_to_delete(
            self,
            interval_s: int,
    ) -> List[bytes]:
        """ Get squeaks older than retention that meet the
        criteria for deletion.
        """
        s = (
            select([self.squeaks, self.profiles])
            .select_from(
                self.squeaks.outerjoin(
                    self.profiles,
                    self.profiles.c.address == self.squeaks.c.author_address,
                )
            )
            .where(self.squeak_is_older_than_retention(interval_s))
            .where(not_(self.profile_has_private_key))
            .where(not_(self.squeak_is_liked))
        )
        with self.get_connection() as connection:
            result = connection.execute(s)
            rows = result.fetchall()
            hashes = [(row["hash"]) for row in rows]
            return hashes

    def insert_profile(self, squeak_profile: SqueakProfile) -> int:
        """ Insert a new squeak profile. """
        ins = self.profiles.insert().values(
            created_time_ms=self.timestamp_now_ms,
            profile_name=squeak_profile.profile_name,
            private_key=squeak_profile.private_key,
            address=squeak_profile.address,
            following=squeak_profile.following,
        )
        with self.get_connection() as connection:
            res = connection.execute(ins)
            profile_id = res.inserted_primary_key[0]
            return profile_id

    def get_profiles(self) -> List[SqueakProfile]:
        """ Get all profiles. """
        s = select([self.profiles])
        with self.get_connection() as connection:
            result = connection.execute(s)
            rows = result.fetchall()
            profiles = [self._parse_squeak_profile(row) for row in rows]
            return profiles

    def get_signing_profiles(self) -> List[SqueakProfile]:
        """ Get all signing profiles. """
        s = select([self.profiles]).where(self.profile_has_private_key)
        with self.get_connection() as connection:
            result = connection.execute(s)
            rows = result.fetchall()
            profiles = [self._parse_squeak_profile(row) for row in rows]
            return profiles

    def get_contact_profiles(self) -> List[SqueakProfile]:
        """ Get all contact profiles. """
        s = select([self.profiles]).where(not_(self.profile_has_private_key))
        with self.get_connection() as connection:
            result = connection.execute(s)
            rows = result.fetchall()
            profiles = [self._parse_squeak_profile(row) for row in rows]
            return profiles

    def get_following_profiles(self) -> List[SqueakProfile]:
        """ Get all following profiles. """
        s = select([self.profiles]).where(self.profiles.c.following)
        with self.get_connection() as connection:
            result = connection.execute(s)
            rows = result.fetchall()
            profiles = [self._parse_squeak_profile(row) for row in rows]
            return profiles

    def get_profile(self, profile_id: int) -> Optional[SqueakProfile]:
        """ Get a profile. """
        s = select([self.profiles]).where(
            self.profiles.c.profile_id == profile_id)
        with self.get_connection() as connection:
            result = connection.execute(s)
            row = result.fetchone()
            if row is None:
                return None
            return self._parse_squeak_profile(row)

    def get_profile_by_address(self, address: str) -> Optional[SqueakProfile]:
        """ Get a profile by address. """
        s = select([self.profiles]).where(self.profiles.c.address == address)
        with self.get_connection() as connection:
            result = connection.execute(s)
            row = result.fetchone()
            if row is None:
                return None
            return self._parse_squeak_profile(row)

    def get_profile_by_name(self, name: str) -> Optional[SqueakProfile]:
        """ Get a profile by name. """
        s = select([self.profiles]).where(self.profiles.c.profile_name == name)
        with self.get_connection() as connection:
            result = connection.execute(s)
            row = result.fetchone()
            if row is None:
                return None
            return self._parse_squeak_profile(row)

    def set_profile_following(self, profile_id: int, following: bool) -> None:
        """ Set a profile is following. """
        stmt = (
            self.profiles.update()
            .where(self.profiles.c.profile_id == profile_id)
            .values(following=following)
        )
        with self.get_connection() as connection:
            connection.execute(stmt)

    def set_profile_name(self, profile_id: int, profile_name: str) -> None:
        """ Set a profile name. """
        stmt = (
            self.profiles.update()
            .where(self.profiles.c.profile_id == profile_id)
            .values(profile_name=profile_name)
        )
        with self.get_connection() as connection:
            connection.execute(stmt)

    def delete_profile(self, profile_id: int) -> None:
        """ Delete a profile. """
        delete_profile_stmt = self.profiles.delete().where(
            self.profiles.c.profile_id == profile_id
        )
        with self.get_connection() as connection:
            connection.execute(delete_profile_stmt)

    def set_profile_image(self, profile_id: int, profile_image: bytes) -> None:
        """ Set a profile image. """
        stmt = (
            self.profiles.update()
            .where(self.profiles.c.profile_id == profile_id)
            .values(profile_image=profile_image)
        )
        with self.get_connection() as connection:
            connection.execute(stmt)

    def set_squeak_decryption_key(self, squeak_hash: bytes, secret_key: bytes, content: str) -> None:
        """ Set the decryption key and decrypted content of a squeak. """
        stmt = (
            self.squeaks.update()
            .where(self.squeaks.c.hash == squeak_hash)
            .values(
                secret_key=secret_key,
                content=content,
            )
        )
        with self.get_connection() as connection:
            connection.execute(stmt)

    def set_squeak_liked(self, squeak_hash: bytes) -> None:
        """ Set the squeak to be liked. """
        stmt = (
            self.squeaks.update()
            .where(self.squeaks.c.hash == squeak_hash)
            .values(liked_time_ms=self.timestamp_now_ms)
        )
        with self.get_connection() as connection:
            connection.execute(stmt)

    def set_squeak_unliked(self, squeak_hash: bytes) -> None:
        """ Set the squeak to be unliked. """
        stmt = (
            self.squeaks.update()
            .where(self.squeaks.c.hash == squeak_hash)
            .values(liked_time_ms=None)
        )
        with self.get_connection() as connection:
            connection.execute(stmt)

    def delete_squeak(self, squeak_hash: bytes) -> None:
        """ Delete a squeak. """
        delete_squeak_stmt = self.squeaks.delete().where(
            self.squeaks.c.hash == squeak_hash
        )
        with self.get_connection() as connection:
            connection.execute(delete_squeak_stmt)

    def insert_peer(self, squeak_peer: SqueakPeer) -> int:
        """ Insert a new squeak peer. """
        ins = self.peers.insert().values(
            created_time_ms=self.timestamp_now_ms,
            peer_name=squeak_peer.peer_name,
            network=squeak_peer.address.network.name,
            host=squeak_peer.address.host,
            port=squeak_peer.address.port,
            autoconnect=squeak_peer.autoconnect,
            share_for_free=squeak_peer.share_for_free,
        )
        with self.get_connection() as connection:
            res = connection.execute(ins)
            id = res.inserted_primary_key[0]
            return id

    def get_peer(self, peer_id: int) -> Optional[SqueakPeer]:
        """ Get a peer. """
        s = select([self.peers]).where(self.peers.c.peer_id == peer_id)
        with self.get_connection() as connection:
            result = connection.execute(s)
            row = result.fetchone()
            if row is None:
                return None
            return self._parse_squeak_peer(row)

    def get_peer_by_address(self, peer_address: PeerAddress) -> Optional[SqueakPeer]:
        """ Get a peer by address. """
        s = (
            select([self.peers])
            .where(self.peers.c.host == peer_address.host)
            .where(self.peers.c.port == peer_address.port)
        )
        with self.get_connection() as connection:
            result = connection.execute(s)
            row = result.fetchone()
            if row is None:
                return None
            return self._parse_squeak_peer(row)

    def get_peers(self) -> List[SqueakPeer]:
        """ Get all peers. """
        s = select([self.peers])
        with self.get_connection() as connection:
            result = connection.execute(s)
            rows = result.fetchall()
            peers = [self._parse_squeak_peer(row) for row in rows]
            return peers

    def get_autoconnect_peers(self) -> List[SqueakPeer]:
        """ Get peers that are set to be autoconnect. """
        s = select([self.peers]).where(self.peers.c.autoconnect)
        with self.get_connection() as connection:
            result = connection.execute(s)
            rows = result.fetchall()
            peers = [self._parse_squeak_peer(row) for row in rows]
            return peers

    def set_peer_autoconnect(self, peer_id: int, autoconnect: bool):
        """ Set a peer is autoconnect. """
        stmt = (
            self.peers.update()
            .where(self.peers.c.peer_id == peer_id)
            .values(autoconnect=autoconnect)
        )
        with self.get_connection() as connection:
            connection.execute(stmt)

    def set_peer_share_for_free(self, peer_id: int, share_for_free: bool):
        """ Set a peer is share_for_free. """
        stmt = (
            self.peers.update()
            .where(self.peers.c.peer_id == peer_id)
            .values(share_for_free=share_for_free)
        )
        with self.get_connection() as connection:
            connection.execute(stmt)

    def set_peer_name(self, peer_id: int, peer_name: str):
        """ Set a peer name. """
        stmt = (
            self.peers.update()
            .where(self.peers.c.peer_id == peer_id)
            .values(peer_name=peer_name)
        )
        with self.get_connection() as connection:
            connection.execute(stmt)

    def delete_peer(self, peer_id: int):
        """ Delete a peer. """
        delete_peer_stmt = self.peers.delete().where(self.peers.c.peer_id == peer_id)
        with self.get_connection() as connection:
            connection.execute(delete_peer_stmt)

    def insert_received_offer(self, received_offer: ReceivedOffer) -> Optional[int]:
        """ Insert a new received offer.

        Return the received offer id of the inserted received offer.
        Return None if received offer already exists.
        """
        ins = self.received_offers.insert().values(
            created_time_ms=self.timestamp_now_ms,
            squeak_hash=received_offer.squeak_hash,
            payment_hash=received_offer.payment_hash,
            nonce=received_offer.nonce,
            payment_point=received_offer.payment_point,
            invoice_timestamp=received_offer.invoice_timestamp,
            invoice_expiry=received_offer.invoice_expiry,
            price_msat=received_offer.price_msat,
            payment_request=received_offer.payment_request,
            destination=received_offer.destination,
            lightning_host=received_offer.lightning_address.host,
            lightning_port=received_offer.lightning_address.port,
            peer_network=received_offer.peer_address.network.name,
            peer_host=received_offer.peer_address.host,
            peer_port=received_offer.peer_address.port,
        )
        with self.get_connection() as connection:
            try:
                res = connection.execute(ins)
                received_offer_id = res.inserted_primary_key[0]
                return received_offer_id
            except sqlalchemy.exc.IntegrityError:
                logger.debug("Failed to insert received offer.", exc_info=True)
                return None

    def get_received_offers(self, squeak_hash: bytes) -> List[ReceivedOffer]:
        """ Get offers with peer for a squeak hash. """
        s = (
            select([self.received_offers])
            .where(self.received_offers.c.squeak_hash == squeak_hash)
            .where(not_(self.received_offer_is_paid))
            .where(not_(self.received_offer_invoice_is_expired))
        )
        with self.get_connection() as connection:
            result = connection.execute(s)
            rows = result.fetchall()
            offers = [
                self._parse_received_offer(row)
                for row in rows
            ]
            return offers

    def get_received_offer(self, received_offer_id: int) -> Optional[ReceivedOffer]:
        """ Get offer with peer for an offer id. """
        s = (
            select([self.received_offers])
            .where(self.received_offers.c.received_offer_id == received_offer_id)
        )
        with self.get_connection() as connection:
            result = connection.execute(s)
            row = result.fetchone()
            if row is None:
                return None
            offer = self._parse_received_offer(row)
            return offer

    def delete_expired_received_offers(self, interval_s):
        """ Delete all expired offers. """
        s = self.received_offers.delete().where(
            or_(
                self.received_offer_invoice_is_expired,
                self.received_offer_is_out_of_retention(interval_s),
            )
        )
        with self.get_connection() as connection:
            res = connection.execute(s)
            deleted_offers = res.rowcount
            return deleted_offers

    def set_received_offer_paid(self, payment_hash: bytes, paid: bool) -> None:
        """ Set a received offer is paid. """
        stmt = (
            self.received_offers.update()
            .where(self.received_offers.c.payment_hash == payment_hash)
            .values(paid=paid)
        )
        with self.get_connection() as connection:
            connection.execute(stmt)

    def insert_sent_payment(self, sent_payment: SentPayment):
        """ Insert a new sent payment. """
        ins = self.sent_payments.insert().values(
            created_time_ms=self.timestamp_now_ms,
            peer_network=sent_payment.peer_address.network.name,
            peer_host=sent_payment.peer_address.host,
            peer_port=sent_payment.peer_address.port,
            squeak_hash=sent_payment.squeak_hash,
            payment_hash=sent_payment.payment_hash,
            secret_key=sent_payment.secret_key,
            price_msat=sent_payment.price_msat,
            node_pubkey=sent_payment.node_pubkey,
            valid=sent_payment.valid,
        )
        with self.get_connection() as connection:
            res = connection.execute(ins)
            sent_payment_id = res.inserted_primary_key[0]
            return sent_payment_id

    def get_sent_payments(
            self,
            limit: int,
            last_sent_payment: Optional[SentPayment],
    ) -> List[SentPayment]:
        """ Get all sent payments. """
        last_created_time = last_sent_payment.created_time_ms if last_sent_payment else self.timestamp_now_ms
        last_payment_hash = last_sent_payment.payment_hash if last_sent_payment else MAX_HASH
        logger.info("""Get sent payments db query with
        limit: {}
        last_created_time: {}
        last_payment_hash: {}
        """.format(
            limit,
            last_created_time,
            last_payment_hash.hex(),
        ))
        s = (
            select([self.sent_payments])
            .where(
                tuple_(
                    self.sent_payments.c.created_time_ms,
                    self.sent_payments.c.payment_hash,
                ) < tuple_(
                    last_created_time,
                    last_payment_hash,
                )
            )
            .order_by(
                self.sent_payments.c.created_time_ms.desc(),
                self.sent_payments.c.payment_hash.desc(),
            )
            .limit(limit)
        )
        with self.get_connection() as connection:
            result = connection.execute(s)
            rows = result.fetchall()
            sent_payments = [
                self._parse_sent_payment(row) for row in rows]
            return sent_payments

    def get_sent_payment(self, sent_payment_id: int) -> Optional[SentPayment]:
        """ Get sent payment by id. """
        s = (
            select([self.sent_payments])
            .where(self.sent_payments.c.sent_payment_id == sent_payment_id)
        )
        with self.get_connection() as connection:
            result = connection.execute(s)
            row = result.fetchone()
            if row is None:
                return None
            return self._parse_sent_payment(row)

    def insert_sent_offer(self, sent_offer: SentOffer):
        """ Insert a new sent offer. """
        ins = self.sent_offers.insert().values(
            created_time_ms=self.timestamp_now_ms,
            squeak_hash=sent_offer.squeak_hash,
            payment_hash=sent_offer.payment_hash,
            nonce=sent_offer.nonce,
            price_msat=sent_offer.price_msat,
            payment_request=sent_offer.payment_request,
            invoice_timestamp=sent_offer.invoice_time,
            invoice_expiry=sent_offer.invoice_expiry,
            peer_network=sent_offer.peer_address.network.name,
            peer_host=sent_offer.peer_address.host,
            peer_port=sent_offer.peer_address.port,
        )
        with self.get_connection() as connection:
            res = connection.execute(ins)
            sent_offer_id = res.inserted_primary_key[0]
            return sent_offer_id

    def get_sent_offers(self) -> List[SentOffer]:
        """ Get all received payments. """
        s = select([self.sent_offers]).order_by(
            self.sent_offers.c.created_time_ms.desc(),
        )
        with self.get_connection() as connection:
            result = connection.execute(s)
            rows = result.fetchall()
            sent_offers = [self._parse_sent_offer(row) for row in rows]
            return sent_offers

    def get_sent_offer_by_payment_hash(self, payment_hash: bytes) -> Optional[SentOffer]:
        """ Get a sent offer by preimage hash. """
        s = select([self.sent_offers]).where(
            self.sent_offers.c.payment_hash == payment_hash
        )
        with self.get_connection() as connection:
            result = connection.execute(s)
            row = result.fetchone()
            if row is None:
                return None
            sent_offer = self._parse_sent_offer(row)
            return sent_offer

    def get_sent_offer_by_squeak_hash_and_peer(self, squeak_hash: bytes, peer_address: PeerAddress) -> Optional[SentOffer]:
        """
        Get a sent offer by squeak hash and peer address host. Only
        return sent offer if it's not expired and not paid.

        TODO: add where clause for peer address network.
        """
        s = (
            select([self.sent_offers])
            .where(self.sent_offers.c.squeak_hash == squeak_hash)
            .where(self.sent_offers.c.peer_network == peer_address.network.name)
            .where(self.sent_offers.c.peer_host == peer_address.host)
            .where(not_(self.sent_offer_is_paid))
            .where(not_(self.sent_offer_is_expired))
        )
        with self.get_connection() as connection:
            result = connection.execute(s)
            row = result.fetchone()
            if row is None:
                return None
            sent_offer = self._parse_sent_offer(row)
            return sent_offer

    def delete_expired_sent_offers(self, interval_s):
        """
        Delete all expired sent offers. Only delete sent
        offers that have been expired for more than interval_s seconds.
        """
        s = self.sent_offers.delete().where(
            self.sent_offer_out_of_retention(interval_s)
        )
        with self.get_connection() as connection:
            res = connection.execute(s)
            deleted_sent_offers = res.rowcount
            return deleted_sent_offers

    def set_sent_offer_paid(self, payment_hash: bytes, paid: bool) -> None:
        """ Set a sent offer is paid. """
        stmt = (
            self.sent_offers.update()
            .where(self.sent_offers.c.payment_hash == payment_hash)
            .values(paid=paid)
        )
        with self.get_connection() as connection:
            connection.execute(stmt)

    def insert_received_payment(self, received_payment: ReceivedPayment) -> Optional[int]:
        """ Insert a new received payment.

        Return the received payment id of the inserted received payment.
        Return None if received payment already exists.
        """
        ins = self.received_payments.insert().values(
            created_time_ms=self.timestamp_now_ms,
            squeak_hash=received_payment.squeak_hash,
            payment_hash=received_payment.payment_hash,
            price_msat=received_payment.price_msat,
            settle_index=received_payment.settle_index,
            peer_network=received_payment.peer_address.network.name,
            peer_host=received_payment.peer_address.host,
            peer_port=received_payment.peer_address.port,
        )
        with self.get_connection() as connection:
            try:
                res = connection.execute(ins)
                received_payment_id = res.inserted_primary_key[0]
                return received_payment_id
            except sqlalchemy.exc.IntegrityError:
                logger.debug(
                    "Failed to insert received payment.", exc_info=True)
                return None

    def get_received_payments(
            self,
            limit: int,
            last_received_payment: Optional[ReceivedPayment],
    ) -> List[ReceivedPayment]:
        """ Get all received payments. """
        last_created_time = last_received_payment.created_time_ms if last_received_payment else self.timestamp_now_ms
        last_payment_hash = last_received_payment.payment_hash if last_received_payment else MAX_HASH
        logger.info("""Get received payments db query with
        limit: {}
        last_created_time: {}
        last_payment_hash: {}
        """.format(
            limit,
            last_created_time,
            last_payment_hash.hex(),
        ))
        s = (
            select([self.received_payments])
            .order_by(
            )
            .where(
                tuple_(
                    self.received_payments.c.created_time_ms,
                    self.received_payments.c.payment_hash,
                ) < tuple_(
                    last_created_time,
                    last_payment_hash,
                )
            )
            .order_by(
                self.received_payments.c.created_time_ms.desc(),
                self.received_payments.c.payment_hash.desc(),
            )
            .limit(limit)
        )
        with self.get_connection() as connection:
            result = connection.execute(s)
            rows = result.fetchall()
            received_payments = [
                self._parse_received_payment(row) for row in rows]
            return received_payments

    def get_latest_settle_index(self) -> Optional[int]:
        """ Get the lnd settled index of the most recent received payment. """
        s = select(
            [func.max(self.received_payments.c.settle_index)],
        ).select_from(self.received_payments)
        with self.get_connection() as connection:
            result = connection.execute(s)
            row = result.fetchone()
            latest_index = row[0]
            return latest_index

    def clear_received_payment_settle_indices(self) -> None:
        """ Set settle_index to zero for all received payments. """
        stmt = (
            self.received_payments.update()
            .values(settle_index=0)
        )
        with self.get_connection() as connection:
            connection.execute(stmt)

    def yield_received_payments_from_index(self, start_index: int = 0) -> Iterator[ReceivedPayment]:
        """ Get all received payments. """
        s = (
            select([self.received_payments])
            .order_by(
                self.received_payments.c.received_payment_id.asc(),
            )
            .where(self.received_payments.c.received_payment_id > start_index)
        )
        with self.get_connection() as connection:
            result = connection.execute(s)
            for row in result:
                received_payment = self._parse_received_payment(row)
                yield received_payment

    def get_received_payment_summary(self) -> ReceivedPaymentSummary:
        """ Get received payment summary. """
        s = select([
            func.count().label("num_payments_received"),
            func.sum(self.received_payments.c.price_msat).label(
                "total_amount_received_msat"),
        ]).select_from(self.received_payments)
        with self.get_connection() as connection:
            result = connection.execute(s)
            row = result.fetchone()
            received_payment_summary = self._parse_received_payment_summary(
                row)
            return received_payment_summary

    def get_sent_payment_summary(self) -> SentPaymentSummary:
        """ Get sent payment summary. """
        s = select([
            func.count().label("num_payments_sent"),
            func.sum(self.sent_payments.c.price_msat).label(
                "total_amount_sent_msat"),
        ]).select_from(self.sent_payments)
        with self.get_connection() as connection:
            result = connection.execute(s)
            row = result.fetchone()
            sent_payment_summary = self._parse_sent_payment_summary(row)
            return sent_payment_summary

    def insert_config(self, user_config: UserConfig) -> Optional[str]:
        """ Insert a new config.

        Return the name (str) of the inserted config user.
        Return None if config already exists.
        """
        ins = self.configs.insert().values(
            username=user_config.username,
            twitter_bearer_token=user_config.twitter_bearer_token,
        )
        with self.get_connection() as connection:
            try:
                res = connection.execute(ins)
                username = res.inserted_primary_key[0]
                return username
            except sqlalchemy.exc.IntegrityError:
                logger.debug("Failed to insert config.", exc_info=True)
                return None

    def get_config(self, username: str) -> Optional[UserConfig]:
        """ Get a config. """
        s = select([self.configs]).where(self.configs.c.username == username)
        with self.get_connection() as connection:
            result = connection.execute(s)
            row = result.fetchone()
            if row is None:
                return None
            return self._parse_user_config(row)

    def set_config_twitter_bearer_token(self, username: str, twitter_bearer_token: str) -> None:
        """ Set a config twitter bearer token. """
        stmt = (
            self.configs.update()
            .where(self.configs.c.username == username)
            .values(twitter_bearer_token=twitter_bearer_token)
        )
        with self.get_connection() as connection:
            connection.execute(stmt)

    def set_config_sell_price_msat(self, username: str, sell_price_msat: int) -> None:
        """ Set a config sell price msat. """
        stmt = (
            self.configs.update()
            .where(self.configs.c.username == username)
            .values(sell_price_msat=sell_price_msat)
        )
        with self.get_connection() as connection:
            connection.execute(stmt)

    def clear_config_sell_price_msat(self, username: str) -> None:
        """ Clear a config sell price msat. """
        stmt = (
            self.configs.update()
            .where(self.configs.c.username == username)
            .values(sell_price_msat=None)
        )
        with self.get_connection() as connection:
            connection.execute(stmt)

    def insert_twitter_account(self, twitter_account: TwitterAccount) -> Optional[int]:
        """ Insert a new twitter account mapping to a squeak profile.

        Return the id (int) of the inserted twitter account.
        Return None if twitter account already exists.
        """
        ins = self.twitter_accounts.insert().values(
            handle=twitter_account.handle,
            profile_id=twitter_account.profile_id,
        )
        with self.get_connection() as connection:
            try:
                res = connection.execute(ins)
                twitter_account_id = res.inserted_primary_key[0]
                return twitter_account_id
            except sqlalchemy.exc.IntegrityError:
                logger.debug(
                    "Failed to insert twitter account.", exc_info=True)
                return None

    def get_twitter_accounts(self) -> List[TwitterAccountEntry]:
        """ Get all twitter accounts. """
        s = (
            select([self.twitter_accounts, self.profiles])
            .select_from(
                self.twitter_accounts.outerjoin(
                    self.profiles,
                    self.profiles.c.profile_id == self.twitter_accounts.c.profile_id,
                )
            )
        )
        with self.get_connection() as connection:
            result = connection.execute(s)
            rows = result.fetchall()
            twitter_accounts = [
                self._parse_twitter_account_entry(row) for row in rows]
            return twitter_accounts

    def delete_twitter_account(self, twitter_account_id: int) -> None:
        """ Delete a twitter_account. """
        delete_twitter_account_stmt = self.twitter_accounts.delete().where(
            self.twitter_accounts.c.twitter_account_id == twitter_account_id
        )
        with self.get_connection() as connection:
            connection.execute(delete_twitter_account_stmt)

    def _parse_squeak(self, row) -> CSqueak:
        return CSqueak.deserialize(row["squeak"])

    def _parse_squeak_entry(self, row) -> SqueakEntry:
        secret_key_column = row["secret_key"]
        is_locked = bool(secret_key_column)
        reply_to = (
            row["hash_reply_sqk"]) if row["hash_reply_sqk"] else None
        liked_time_ms = row["liked_time_ms"]
        profile = self._try_parse_squeak_profile(row)
        return SqueakEntry(
            squeak_hash=(row["hash"]),
            address=row["author_address"],
            block_height=row["n_block_height"],
            block_hash=(row["hash_block"]),
            block_time=row["block_time"],
            squeak_time=row["n_time"],
            reply_to=reply_to,
            is_unlocked=is_locked,
            liked_time_ms=liked_time_ms,
            content=row["content"],
            squeak_profile=profile,
        )

    def _parse_squeak_profile(self, row) -> SqueakProfile:
        private_key_column = row["private_key"]
        private_key = bytes(private_key_column) if private_key_column else None
        return SqueakProfile(
            profile_id=row["profile_id"],
            profile_name=row["profile_name"],
            private_key=private_key,
            address=row["address"],
            following=row["following"],
            profile_image=row["profile_image"],
        )

    def _try_parse_squeak_profile(self, row) -> Optional[SqueakProfile]:
        if row["profile_id"] is None:
            return None
        return self._parse_squeak_profile(row)

    def _parse_squeak_peer(self, row) -> SqueakPeer:
        return SqueakPeer(
            peer_id=row[self.peers.c.peer_id],
            peer_name=row["peer_name"],
            address=PeerAddress(
                network=Network[row["network"]],
                host=row["host"],
                port=row["port"],
            ),
            autoconnect=row["autoconnect"],
            share_for_free=row["share_for_free"],
        )

    def _parse_received_offer(self, row) -> ReceivedOffer:
        return ReceivedOffer(
            received_offer_id=row["received_offer_id"],
            squeak_hash=(row["squeak_hash"]),
            payment_hash=(row["payment_hash"]),
            nonce=(row["nonce"]),
            payment_point=(row["payment_point"]),
            invoice_timestamp=row["invoice_timestamp"],
            invoice_expiry=row["invoice_expiry"],
            price_msat=row["price_msat"],
            payment_request=row["payment_request"],
            destination=row["destination"],
            lightning_address=LightningAddressHostPort(
                host=row["lightning_host"],
                port=row["lightning_port"],
            ),
            peer_address=PeerAddress(
                network=Network[row["peer_network"]],
                host=row["peer_host"],
                port=row["peer_port"],
            ),
            paid=row["paid"],
        )

    def _parse_sent_payment(self, row) -> SentPayment:
        return SentPayment(
            sent_payment_id=row["sent_payment_id"],
            created_time_ms=row[self.sent_payments.c.created_time_ms],
            peer_address=PeerAddress(
                network=Network[row["peer_network"]],
                host=row["peer_host"],
                port=row["peer_port"],
            ),
            squeak_hash=(row["squeak_hash"]),
            payment_hash=(row["payment_hash"]),
            secret_key=(row["secret_key"]),
            price_msat=row["price_msat"],
            node_pubkey=row["node_pubkey"],
            valid=row["valid"],
        )

    def _parse_sent_offer(self, row) -> SentOffer:
        return SentOffer(
            sent_offer_id=row["sent_offer_id"],
            squeak_hash=(row["squeak_hash"]),
            payment_hash=(row["payment_hash"]),
            nonce=(row["nonce"]),
            price_msat=row["price_msat"],
            payment_request=row["payment_request"],
            invoice_time=row["invoice_timestamp"],
            invoice_expiry=row["invoice_expiry"],
            peer_address=PeerAddress(
                network=Network[row["peer_network"]],
                host=row["peer_host"],
                port=row["peer_port"],
            ),
            paid=row["paid"],
        )

    def _parse_received_payment(self, row) -> ReceivedPayment:
        return ReceivedPayment(
            received_payment_id=row["received_payment_id"],
            created_time_ms=row["created_time_ms"],
            squeak_hash=(row["squeak_hash"]),
            payment_hash=(row["payment_hash"]),
            price_msat=row["price_msat"],
            settle_index=row["settle_index"],
            peer_address=PeerAddress(
                network=Network[row["peer_network"]],
                host=row["peer_host"],
                port=row["peer_port"],
            ),
        )

    def _parse_received_payment_summary(self, row) -> ReceivedPaymentSummary:
        return ReceivedPaymentSummary(
            num_received_payments=row["num_payments_received"],
            total_amount_received_msat=row["total_amount_received_msat"],
        )

    def _parse_sent_payment_summary(self, row) -> SentPaymentSummary:
        return SentPaymentSummary(
            num_sent_payments=row["num_payments_sent"],
            total_amount_sent_msat=row["total_amount_sent_msat"],
        )

    def _parse_user_config(self, row) -> UserConfig:
        return UserConfig(
            username=row["username"],
            twitter_bearer_token=row["twitter_bearer_token"],
            sell_price_msat=row["sell_price_msat"],
        )

    def _parse_twitter_account_entry(self, row) -> TwitterAccountEntry:
        profile = self._try_parse_squeak_profile(row)
        return TwitterAccountEntry(
            twitter_account_id=row["twitter_account_id"],
            handle=row["handle"],
            profile_id=row["profile_id"],
            profile=profile,
        )



# MIT License
#
# Copyright (c) 2020 Jonathan Zernik
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
import socket
import threading

from squeaknode.core.peer_address import Network
from squeaknode.core.peer_address import PeerAddress


MIN_PEERS = 5
MAX_PEERS = 10
UPDATE_THREAD_SLEEP_TIME = 10


logger = logging.getLogger(__name__)


class PeerServer(object):
    """Accepts incoming connections from other peers in the network.
    """

    def __init__(self, peer_handler, port):
        self.peer_handler = peer_handler
        self.port = port
        self.listen_socket = socket.socket()

    def start(self):
        logger.info("Starting peer server with port: {}".format(
            self.port,
        ))
        # Start Listen thread
        threading.Thread(
            target=self.accept_connections,
            name="peer_server_listen_thread",
        ).start()

    def stop(self):
        logger.info("Stopping peer server listener thread...")
        self.listen_socket.shutdown(socket.SHUT_RDWR)
        self.listen_socket.close()

    def accept_connections(self):
        try:
            self.listen_socket.bind(('', self.port))
            self.listen_socket.listen()
            while True:
                peer_socket, address = self.listen_socket.accept()
                host, port = address
                peer_address = PeerAddress(
                    network=Network.IPV4,
                    host=host,
                    port=port,
                )
                peer_socket.setblocking(True)
                self.handle_connection(
                    peer_socket,
                    peer_address,
                )
        except Exception:
            logger.exception("Accept peer connections failed.")

    def handle_connection(
            self,
            peer_socket: socket.socket,
            peer_address: PeerAddress,
    ):
        """Handle a newly connected peer socket."""
        threading.Thread(
            target=self.peer_handler.handle_connection,
            args=(peer_socket, peer_address, False),
        ).start()



# MIT License
#
# Copyright (c) 2020 Jonathan Zernik
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import pytest
from bitcoin.core import CBlockHeader
from squeak.core.elliptic import payment_point_bytes_from_scalar_bytes
from squeak.core.signing import CSigningKey
from squeak.core.signing import CSqueakAddress

from squeaknode.bitcoin.block_info import BlockInfo
from squeaknode.core.connected_peer import ConnectedPeer
from squeaknode.core.download_result import DownloadResult
from squeaknode.core.lightning_address import LightningAddressHostPort
from squeaknode.core.offer import Offer
from squeaknode.core.peer_address import Network
from squeaknode.core.peer_address import PeerAddress
from squeaknode.core.received_offer import ReceivedOffer
from squeaknode.core.received_payment import ReceivedPayment
from squeaknode.core.received_payment_summary import ReceivedPaymentSummary
from squeaknode.core.secret_keys import add_tweak
from squeaknode.core.secret_keys import generate_tweak
from squeaknode.core.sent_offer import SentOffer
from squeaknode.core.sent_payment import SentPayment
from squeaknode.core.sent_payment_summary import SentPaymentSummary
from squeaknode.core.squeak_entry import SqueakEntry
from squeaknode.core.squeak_peer import SqueakPeer
from squeaknode.core.squeaks import get_hash
from squeaknode.core.squeaks import make_squeak_with_block
from squeaknode.core.user_config import UserConfig
from squeaknode.network.peer import Peer
from tests.utils import gen_contact_profile
from tests.utils import gen_signing_profile
from tests.utils import sha256


@pytest.fixture
def signing_key():
    yield CSigningKey.generate()


@pytest.fixture
def signing_key_bytes(signing_key):
    yield str(signing_key).encode()


@pytest.fixture
def address(signing_key):
    verifying_key = signing_key.get_verifying_key()
    yield CSqueakAddress.from_verifying_key(verifying_key)


@pytest.fixture
def address_str(address):
    yield str(address)


@pytest.fixture
def block_count():
    yield 555


@pytest.fixture
def block_hash_str():
    # Block hash of block at height 555
    yield '00000000edade40797e3c4bf27edeb65733d1884beaa8c502a89d50a54111e1c'


@pytest.fixture
def block_hash(block_hash_str):
    yield bytes.fromhex(block_hash_str)


@pytest.fixture
def block_header_str():
    # Block header of block at height 555
    yield '0100000079c30d2c23727a1e9f5feda4e7feb8ea0bda2ab98e23e7f6a9cf594f00000000b0de897e42fa7a3b5c3a6bfb8e797acf4ffbc16169394b03ad93296524ed633dcfef6e49ffff001d36d19a6c'


@pytest.fixture
def block_header_bytes(block_header_str):
    yield bytes.fromhex(block_header_str)


@pytest.fixture
def block_header(block_header_bytes):
    yield CBlockHeader.deserialize(block_header_bytes)


@pytest.fixture
def block_time(block_header):
    yield block_header.nTime


@pytest.fixture
def block_info(block_count, block_hash, block_header):
    yield BlockInfo(
        block_height=block_count,
        block_hash=block_hash,
        block_header=block_header,
    )


@pytest.fixture
def squeak_content():
    yield "hello!"


@pytest.fixture
def reply_squeak_content():
    yield "this is a reply!"


@pytest.fixture
def squeak_and_secret_key(signing_key, squeak_content, block_info):
    yield make_squeak_with_block(
        signing_key,
        squeak_content,
        block_info.block_height,
        block_info.block_hash,
    )


@pytest.fixture
def reply_squeak_and_secret_key(signing_key, reply_squeak_content, block_info, squeak_hash):
    yield make_squeak_with_block(
        signing_key,
        reply_squeak_content,
        block_info.block_height,
        block_info.block_hash,
        replyto_hash=squeak_hash,
    )


@pytest.fixture
def squeak(squeak_and_secret_key):
    squeak, _ = squeak_and_secret_key
    yield squeak


@pytest.fixture
def secret_key(squeak_and_secret_key):
    _, secret_key = squeak_and_secret_key
    yield secret_key


@pytest.fixture
def squeak_hash(squeak):
    yield get_hash(squeak)


@pytest.fixture
def squeak_hash_str(squeak_hash):
    yield squeak_hash.hex()


@pytest.fixture
def squeak_bytes(squeak):
    yield squeak.serialize()


@pytest.fixture
def squeak_time(squeak):
    yield squeak.nTime


@pytest.fixture
def squeak_reply_to_hash(squeak):
    yield None


@pytest.fixture
def reply_squeak(reply_squeak_and_secret_key):
    squeak, _ = reply_squeak_and_secret_key
    yield squeak


@pytest.fixture
def reply_squeak_hash(reply_squeak):
    yield get_hash(reply_squeak)


@pytest.fixture
def peer_address():
    yield PeerAddress(
        network=Network.IPV4,
        host="fake_host",
        port=8765,
    )


@pytest.fixture
def signing_profile_name():
    yield "fake_signing_profile_name"


@pytest.fixture
def contact_profile_name():
    yield "fake_contact_profile_name"


@pytest.fixture
def signing_profile(signing_profile_name, signing_key):
    yield gen_signing_profile(
        signing_profile_name,
        str(signing_key),
    )


@pytest.fixture
def contact_profile(contact_profile_name, address):
    yield gen_contact_profile(
        contact_profile_name,
        str(address),
    )


@pytest.fixture
def squeak_entry_locked(
        squeak,
        squeak_hash,
        address_str,
        block_count,
        block_hash,
        block_time,
        squeak_time,
        squeak_reply_to_hash,
        signing_profile,
):
    yield SqueakEntry(
        squeak_hash=squeak_hash,
        address=address_str,
        block_height=block_count,
        block_hash=block_hash,
        block_time=block_time,
        squeak_time=squeak_time,
        reply_to=squeak_reply_to_hash,
        is_unlocked=False,
        squeak_profile=signing_profile,
        liked_time_ms=None,
        content=None,
    )


@pytest.fixture
def peer_name():
    yield "fake_peer_name"


@pytest.fixture
def peer(peer_name, peer_address):
    yield SqueakPeer(
        peer_id=None,
        peer_name=peer_name,
        address=peer_address,
        autoconnect=False,
        share_for_free=False,
    )


@pytest.fixture
def lightning_address():
    return LightningAddressHostPort(host="my_lightning_host", port=8765)


@pytest.fixture
def external_lightning_address():
    return LightningAddressHostPort(host="my_external_lightning_host", port=13579)


@pytest.fixture
def price_msat():
    return 777


@pytest.fixture
def nonce():
    yield generate_tweak()


@pytest.fixture
def preimage(secret_key, nonce):
    yield add_tweak(secret_key, nonce)


@pytest.fixture
def payment_point(secret_key):
    yield payment_point_bytes_from_scalar_bytes(secret_key)


@pytest.fixture
def payment_hash(preimage):
    # TODO: When PTLC is used, this should be the payment point of preimage.
    yield sha256(preimage)


@pytest.fixture
def payment_request():
    yield "fake_payment_request"


@pytest.fixture
def creation_date():
    yield 777777


@pytest.fixture
def expiry():
    yield 5555


@pytest.fixture
def seller_pubkey():
    yield "fake_seller_pubkey"


@pytest.fixture
def uris(seller_pubkey, lightning_address):
    yield [
        '{}@{}:{}'.format(
            seller_pubkey,
            lightning_address.host,
            lightning_address.port,
        ),
        'fake_pubkey@foobar.com:12345',
        'fake_pubkey@fakehost.com:56789',
    ]


@pytest.fixture
def settle_index():
    yield 345


@pytest.fixture
def sent_offer(
        squeak_hash,
        price_msat,
        payment_hash,
        nonce,
        creation_date,
        expiry,
        payment_request,
        peer_address,
):
    yield SentOffer(
        sent_offer_id=None,
        squeak_hash=squeak_hash,
        payment_hash=payment_hash,
        nonce=nonce,
        price_msat=price_msat,
        payment_request=payment_request,
        invoice_time=creation_date,
        invoice_expiry=expiry,
        peer_address=peer_address,
        paid=False,
    )


@pytest.fixture
def received_offer(
        squeak_hash,
        price_msat,
        payment_hash,
        nonce,
        payment_point,
        creation_date,
        expiry,
        payment_request,
        seller_pubkey,
        lightning_address,
        peer_address,
):
    yield ReceivedOffer(
        received_offer_id=None,
        squeak_hash=squeak_hash,
        price_msat=price_msat,
        payment_hash=payment_hash,
        nonce=nonce,
        payment_point=payment_point,
        invoice_timestamp=creation_date,
        invoice_expiry=expiry,
        payment_request=payment_request,
        destination=seller_pubkey,
        lightning_address=lightning_address,
        peer_address=peer_address,
        paid=False,
    )


@pytest.fixture
def received_payment(
        squeak_hash,
        payment_hash,
        price_msat,
        settle_index,
        peer_address,
):
    yield ReceivedPayment(
        received_payment_id=None,
        created_time_ms=None,
        squeak_hash=squeak_hash,
        payment_hash=payment_hash,
        price_msat=price_msat,
        settle_index=settle_index,
        peer_address=peer_address,
    )


@pytest.fixture
def sent_payment(
        squeak_hash,
        payment_hash,
        secret_key,
        price_msat,
        seller_pubkey,
        peer_address,
):
    yield SentPayment(
        sent_payment_id=None,
        created_time_ms=None,
        peer_address=peer_address,
        squeak_hash=squeak_hash,
        payment_hash=payment_hash,
        secret_key=secret_key,
        price_msat=price_msat,
        node_pubkey=seller_pubkey,
        valid=True,
    )


@pytest.fixture
def offer(
        squeak_hash,
        nonce,
        payment_request,
        lightning_address,
):
    yield Offer(
        squeak_hash=squeak_hash,
        nonce=nonce,
        payment_request=payment_request,
        host=lightning_address.host,
        port=lightning_address.port,
    )


@pytest.fixture
def num_received_payments():
    yield 574


@pytest.fixture
def num_sent_payments():
    yield 189


@pytest.fixture
def total_amount_received_msat():
    yield 3259084


@pytest.fixture
def total_amount_sent_msat():
    yield 198374


@pytest.fixture
def received_payment_summary(
        num_received_payments,
        total_amount_received_msat,
):
    yield ReceivedPaymentSummary(
        num_received_payments=num_received_payments,
        total_amount_received_msat=total_amount_received_msat,
    )


@pytest.fixture
def sent_payment_summary(
        num_sent_payments,
        total_amount_sent_msat,
):
    yield SentPaymentSummary(
        num_sent_payments=num_sent_payments,
        total_amount_sent_msat=total_amount_sent_msat,
    )


@pytest.fixture
def peer_object(peer_address):
    yield Peer(
        peer_socket=None,
        local_address=peer_address,
        remote_address=peer_address,
        outgoing=True,
        peer_changed_listener=None,
    )


@pytest.fixture
def connected_peer(peer, peer_object):
    yield ConnectedPeer(
        peer=peer_object,
        saved_peer=peer,
    )


@pytest.fixture
def download_result():
    yield DownloadResult(
        number_downloaded=1,
        number_requested=20,
        elapsed_time_ms=56789,
        number_peers=4,
    )


@pytest.fixture
def user_config():
    yield UserConfig(
        username="default_user",
    )


@pytest.fixture
def twitter_bearer_token():
    yield 'abcdefg987654321'



# MIT License
#
# Copyright (c) 2020 Jonathan Zernik
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
import queue
import threading
from contextlib import contextmanager

logger = logging.getLogger(__name__)

DEFAULT_MAX_QUEUE_SIZE = 1000
DEFAULT_UPDATE_INTERVAL_S = 1


class ReceivedPaymentsSubscriptionClient:
    """Class that can be used to get a subscription to a stream
    of received payments.

    `initial_index` parameter refers to the `received_payment_id` that will be
    the first in the result stream.
    """

    def __init__(
        self,
        squeak_db,
        initial_index: int,
        stopped: threading.Event,
        max_queue_size=DEFAULT_MAX_QUEUE_SIZE,
        update_interval_s=DEFAULT_UPDATE_INTERVAL_S,
    ):
        self.squeak_db = squeak_db
        self.initial_index = initial_index
        self.stopped = stopped
        self.update_interval_s = update_interval_s
        self.q: queue.Queue = queue.Queue(max_queue_size)

    @contextmanager
    def open_subscription(self):
        # Start the thread to populate the queue.
        threading.Thread(
            target=self.populate_queue,
        ).start()
        try:
            logger.info("Before yielding received payment client...")
            yield self
            logger.info("After yielding received payment client...")
        finally:
            logger.info("Stopping received payment client...")
            self.stopped.set()
            logger.info("Stopped received payment client...")

    def populate_queue(self):
        payment_index = self.initial_index
        while not self.stopped.is_set():
            try:
                for payment in self.get_latest_received_payments(payment_index):
                    self.q.put(payment)
                    payment_index = payment.received_payment_id
                    logger.info(
                        "Added payment to queue. Size: {}".format(
                            self.q.qsize())
                    )
            except Exception:
                logger.error(
                    "Exception while populating queue.",
                    exc_info=True,
                )
            self.stopped.wait(self.update_interval_s)
        # Put the poison pill
        logger.info("Putting poison pill in queue...")
        self.q.put(None)

    def get_latest_received_payments(self, payment_index):
        return self.squeak_db.yield_received_payments_from_index(
            payment_index,
        )

    def get_received_payments(self):
        while True:
            payment = self.q.get()
            if payment is None:
                logger.debug("Poison pill swallowed.")
                return
            yield payment
            self.q.task_done()
            logger.info(
                "Removed payment from queue. Size: {}".format(
                    self.q.qsize())
            )
